# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_baidu_lian_xi_qu_dian_nao_pei_xun_1643390077():
    verify_serp_parsing(
        "https://web.archive.org/web/20220128171437id_/http://zhannei.baidu.com/cse/site?q=%E6%BF%82%E6%BA%AA%E5%8C%BA%E7%94%B5%E8%84%91%E5%9F%B9%E8%AE%AD&p=0&cc=actwhite.cn",
        "baidu",
    )


def test_parse_query_baidu_mao_kui_lu_mao_cong_cong_mao_ning_kui_mao_dan_dan_mang_lu_1549835769():
    verify_serp_parsing(
        "https://web.archive.org/web/20190210215609id_/https://www.baidu.com/s?wd=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&oq=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&ie=utf-8&rsv_pq=87216f54001538da&rsv_t=19e3cg%2FMtz92Yq09CCIJaki1%2BUL6aGQ6QJQBgVzprgVnaNRLdUeIL6teIzY&rqlang=cn&nojs=1&bqid=87216f54001538da",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1569971117():
    verify_serp_parsing(
        "https://web.archive.org/web/20191001230517id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1544443888():
    verify_serp_parsing(
        "https://web.archive.org/web/20181210121128id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_mao_kui_lu_mao_cong_cong_mao_ning_kui_mao_dan_dan_mang_lu_1564025418():
    verify_serp_parsing(
        "https://web.archive.org/web/20190725033018id_/https://www.baidu.com/s?wd=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&oq=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&ie=utf-8&rsv_pq=d25fbc8d000b4e82&rsv_t=fc63O4O%2FKyZwE5v5jDt8lv18HT8dHQWj2Go%2BcrYpMDHze7HivVjCvAF%2F6R0&rqlang=cn&nojs=1&bqid=d25fbc8d000b4e82",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1547510808():
    verify_serp_parsing(
        "https://web.archive.org/web/20190115000648id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1567878230():
    verify_serp_parsing(
        "https://web.archive.org/web/20190907174350id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_mao_kui_lu_mao_cong_cong_mao_ning_kui_mao_dan_dan_mang_lu_1553838201():
    verify_serp_parsing(
        "https://web.archive.org/web/20190329054321id_/https://www.baidu.com/s?wd=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&oq=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&ie=utf-8&rsv_pq=ce502b7c00017f4e&rsv_t=429fdL0VhHNPGYvy2nJwvedkWEHrVvYOK1YdtSoHeKxoh77YPXsd%2FA%2FeSOo&rqlang=cn&nojs=1&bqid=ce502b7c00017f4e",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1558963051():
    verify_serp_parsing(
        "https://web.archive.org/web/20190527131731id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_zhong_guo_lian_tong_smsc_1372160086():
    verify_serp_parsing(
        "https://web.archive.org/web/20130625113446id_/http://www.baidu.com/s?wd=%E4%B8%AD%E5%9B%BD%E8%81%94%E9%80%9A%20smsc&tn=baidufir",
        "baidu",
    )


def test_parse_query_baidu_sexinsex_1_2_1213858525():
    verify_serp_parsing(
        "https://web.archive.org/web/20080619065525id_/http://www.baidu.com:80/s?wd=sexinsex%B4%F3%C2%BD%C3%E2%B7%D1%C8%EB%BF%DA&lm=0&si=&rn=10&ie=gb2312&ct=0&cl=3&f=1&rsp=9",
        "baidu",
    )


def test_parse_query_baidu_mao_ning_kui_mao_dan_dan_mang_long_lou_mang_lu_shikato_1537858258():
    verify_serp_parsing(
        "https://web.archive.org/web/20180925065058id_/https://www.baidu.com/s?wd=%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E6%B0%93%E9%99%87%E6%90%82%E6%B0%93%E8%B5%82%E8%81%A2&oq=%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E6%B0%93%E9%99%87%E6%90%82%E6%B0%93%E8%B5%82%E8%81%A2&ie=utf-8&rsv_pq=9f20a69b00016e59&rsv_t=4e92hBffxcH%2BZ67tApMjf2jcV67jPo6GzU42zK0wG0%2Ff6aF%2F68rl5AeYfy8&rqlang=cn&nojs=1&bqid=9f20a69b00016e59",
        "baidu",
    )


def test_parse_query_baidu_long_shi_liang_1639014949():
    verify_serp_parsing(
        "https://web.archive.org/web/20211209015549id_/https://wenku.baidu.com/search?lm=0&od=0&ie=utf-8&dyTabStr=MCwzLDQsMSw2LDUsNyw4LDIsOQ%3D%3D&word=%22%E9%BE%8D%E4%B8%96%E6%A8%91%22",
        "baidu",
    )


def test_parse_query_baidu_mao_kui_lu_mao_cong_cong_mao_ning_kui_mao_dan_dan_mang_lu_1558979495():
    verify_serp_parsing(
        "https://web.archive.org/web/20190527175135id_/https://www.baidu.com/s?wd=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&oq=%E8%8C%85%E8%81%B5%E9%A9%B4%E8%8C%85%E8%81%A1%E8%81%A6%E7%8C%AB%E8%81%B9%E8%81%B5%E7%8C%AB%E8%81%B8%E8%81%B8%E5%BF%99%E5%8D%A4&ie=utf-8&rsv_pq=8904574900082530&rsv_t=9bbeaAnORF%2BIrKzJVAMdIO%2B5XlN26aMHDddm0oqJPVHpCwMWT1aVQIDgsHA&rqlang=cn&nojs=1&bqid=8904574900082530",
        "baidu",
    )


def test_parse_query_baidu_eru_1525597474():
    verify_serp_parsing(
        "https://web.archive.org/web/20180506090434id_/http://tieba.baidu.com:80/f?kw=eru&ie=utf-8",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1554431691():
    verify_serp_parsing(
        "https://web.archive.org/web/20190405023451id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_lian_yun_gang_qi_yang_yue_zi_hui_suo_zhong_xin_jing_zhun_ke_hu_ying_xiao_key668_cn_1537259978():
    verify_serp_parsing(
        "https://web.archive.org/web/20180918083938id_/https://www.baidu.com/s?wd=%E8%BF%9E%E4%BA%91%E6%B8%AF%E9%AA%90%E6%89%AC%E6%9C%88%E5%AD%90%E4%BC%9A%E6%89%80%E4%B8%AD%E5%BF%83%E7%B2%BE%E5%87%86%E5%AE%A2%E6%88%B7%E8%90%A5%E9%94%80key668.cn",
        "baidu",
    )


def test_parse_query_baidu_long_feng_qu_qi_quan_qi_huo_pei_xun_1642853767():
    verify_serp_parsing(
        "https://web.archive.org/web/20220122121607id_/http://zhannei.baidu.com/cse/site?q=%E9%BE%99%E5%87%A4%E5%8C%BA%E6%9C%9F%E6%9D%83%E6%9C%9F%E8%B4%A7%E5%9F%B9%E8%AE%AD&p=0&cc=actunderstand.cn",
        "baidu",
    )


def test_parse_query_baidu_e_eeeae_1522165852():
    verify_serp_parsing(
        "https://web.archive.org/web/20180327155052id_/https://www.baidu.com/s?wd=%C3%A9%C2%98%C2%BF%C3%A9%C2%87%C2%8C%C3%A8%C2%9C%C2%98%C3%A8%C2%9B%C2%9B%C3%A6%C2%B1",
        "baidu",
    )


def test_parse_query_baidu_shang_hai_11xuan_5ding_dan_wei_xin_weiwei776699_1545685930():
    verify_serp_parsing(
        "https://web.archive.org/web/20181224211210id_/https://wenku.baidu.com/search?word=%E4%B8%8A%E6%B5%B711%E9%80%895%E5%AE%9A%E8%83%86%2B%E3%80%96%E5%BE%AE%E4%BF%A1weiwei776699%E3%80%97%E2%94%A0&ie=utf-8&lm=0&od=0",
        "baidu",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_etsy_embroidery_kit_1375805089():
    verify_serp_parsing(
        "https://web.archive.org/web/20130806160449id_/http://www.etsy.com/search?includes[]=tags&q=embroidery+kit",
        "etsy",
    )


def test_parse_query_etsy_sundress_1374472455():
    verify_serp_parsing(
        "https://web.archive.org/web/20130722055415id_/http://www.etsy.com:80/search?includes[]=tags&q=Sundress",
        "etsy",
    )


def test_parse_query_etsy_storage_and_organization_1632881953():
    verify_serp_parsing(
        "https://web.archive.org/web/20210929021913id_/https://www.etsy.com/search?q=storage+and+organization&mosv=sese&moci=1040853552759&mosi=1027452046072&ref=hp_bubbles_Fall21_US_GB_CA_FR_DE&anchor_listing_id=830217119",
        "etsy",
    )


def test_parse_query_etsy_andrew_kim_1566153959():
    verify_serp_parsing(
        "https://web.archive.org/web/20190818184559id_/https://www.etsy.com/search?q=Andrew+Kim&explicit=1&order=date_desc",
        "etsy",
    )


def test_parse_query_etsy_pendants_1667648711():
    verify_serp_parsing(
        "https://web.archive.org/web/20221105114511id_/https://www.etsy.com/search?locale_override=USD%7Cen-US%7CUS&explicit=1&q=pendants&anchor_listing_id=200491601&gf_type=interest",
        "etsy",
    )


def test_parse_query_etsy_pacifier_1336437253():
    verify_serp_parsing(
        "https://web.archive.org/web/20120508003413id_/http://www.etsy.com:80/search?includes[]=tags&q=pacifier",
        "etsy",
    )


def test_parse_query_etsy_invitation_1367234841():
    verify_serp_parsing(
        "https://web.archive.org/web/20130429112721id_/http://www.etsy.com/search?q=invitation&ref=exp_listing",
        "etsy",
    )


def test_parse_query_etsy_yan_dong_kuai_le_shi_fen_shou_ji_ban_jia_wei_xin_xw639_2018_1537911471():
    verify_serp_parsing(
        "https://web.archive.org/web/20180925213751id_/https://www.etsy.com/search?q=%E5%B9%BF%E4%B8%9C%E5%BF%AB%E4%B9%90%E5%8D%81%E5%88%86%E6%89%8B%E6%9C%BA%E7%89%88%E5%8A%A0%E5%BE%AE%E3%80%85%E4%BF%A1xw639-2018%E2%99%80",
        "etsy",
    )


def test_parse_query_etsy_hard_plastic_1333819223():
    verify_serp_parsing(
        "https://web.archive.org/web/20120407172023id_/http://www.etsy.com:80/search?includes[]=materials&q=hard+plastic",
        "etsy",
    )


def test_parse_query_etsy_small_animal_1378528595():
    verify_serp_parsing(
        "https://web.archive.org/web/20130907043635id_/http://www.etsy.com/search?includes[]=tags&q=Small+Animal",
        "etsy",
    )



from pathlib import Path

from click import option, Path as PathParam, argument, IntRange, group

from archive_query_log.legacy import DATA_DIRECTORY_PATH, CDX_API_URL
from archive_query_log.legacy.cli.util import URL


@group("alexa")
def alexa():
    pass


@alexa.command("fetch-archived-urls")
@option(
    "-u", "--api-url", "--cdx-api-url",
    type=URL,
    default=CDX_API_URL,
)
@argument(
    "output-path",
    type=PathParam(
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=False,
        resolve_path=True,
        path_type=Path,
    ),
    default=DATA_DIRECTORY_PATH / "alexa-top-1m-archived-urls.jsonl"
)
def archived_urls(api_url: str, output_path: Path) -> None:
    from archive_query_log.legacy.services.alexa import AlexaTop1MArchivedUrls
    AlexaTop1MArchivedUrls(
        output_path=output_path,
        cdx_api_url=api_url,
    ).fetch()


@alexa.command("fuse-domains")
@option(
    "-d", "--data-dir",
    type=PathParam(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        path_type=Path,
    ),
    default=DATA_DIRECTORY_PATH
)
@option(
    "-u", "--api-url", "--cdx-api-url",
    type=URL,
    default=CDX_API_URL,
)
@option(
    "-k", "--depth",
    type=IntRange(min=1),
    default=1000,
)
def domains(data_dir: Path, api_url: str, depth: int) -> None:
    from archive_query_log.legacy.services.alexa import AlexaTop1MFusedDomains
    AlexaTop1MFusedDomains(
        data_directory_path=data_dir,
        cdx_api_url=api_url,
        max_domains_per_ranking=depth,
    ).fetch()



from typing import Mapping

from archive_query_log.legacy import DATA_DIRECTORY_PATH
from archive_query_log.legacy.model import Service
from archive_query_log.legacy.services import read_services

# Load all services that have parsers and create the services for them.
SERVICES_PATH = DATA_DIRECTORY_PATH / "selected-services.yaml"
SERVICES: Mapping[str, Service] = read_services(SERVICES_PATH)



from click import group, option

from archive_query_log.cli.util import pass_config
from archive_query_log.config import Config


@group()
def monitoring() -> None:
    pass


@monitoring.command()
@option("-h", "--host", type=str, default="127.0.0.1",
        help="The interface to bind to.")
@option("-p", "--port", type=int, default=5000,
        help="The port to bind to.")
@pass_config
def run(
        config: Config,
        host: str,
        port: int,
) -> None:
    from archive_query_log.monitoring import run_monitoring
    run_monitoring(config, host, port)



from abc import ABC, abstractmethod
from csv import writer
from dataclasses import dataclass
from functools import cached_property
from gzip import GzipFile
from json import loads, JSONDecodeError
from pathlib import Path
from shelve import open as shelf_open, Shelf  # nosec: B403
from shutil import copyfileobj
from typing import Iterator, TypeVar, Generic, Type, final, \
    ContextManager, Iterable
from uuid import UUID, uuid5, NAMESPACE_URL

from dataclasses_json import DataClassJsonMixin
from marshmallow import Schema
from tqdm.auto import tqdm
from warcio.archiveiterator import ArchiveIterator
from warcio.recordloader import ArcWarcRecord

from archive_query_log.legacy import DATA_DIRECTORY_PATH, LOGGER
from archive_query_log.legacy.model import ArchivedUrl, ArchivedQueryUrl, \
    ArchivedParsedSerp, ArchivedSearchResultSnippet, ArchivedRawSerp, \
    ArchivedRawSearchResult, CorpusJsonlLocation, CorpusJsonlSnippetLocation, \
    CorpusWarcLocation
from archive_query_log.legacy.util.text import count_lines, text_io_wrapper


@dataclass(frozen=True)
class _MetaIndex:
    base_type: str
    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False

    @cached_property
    def base_path(self) -> Path:
        base_path = self.data_directory
        if self.focused:
            base_path /= "focused"
        if self.base_type == "archived-search-result-snippets":
            base_path /= "archived-parsed-serps"
        else:
            base_path /= self.base_type
        return base_path

    def _is_indexable_path(self, path: Path) -> bool:
        if self.base_type == "archived-urls":
            return path.is_file() and path.name.endswith(".jsonl.gz")
        elif self.base_type == "archived-query-urls":
            return path.is_file() and path.name.endswith(".jsonl.gz")
        elif self.base_type == "archived-raw-serps":
            return path.is_dir()
        elif self.base_type == "archived-parsed-serps":
            return path.is_file() and path.name.endswith(".jsonl.gz")
        elif self.base_type == "archived-search-result-snippets":
            return path.is_file() and path.name.endswith(".jsonl.gz")
        elif self.base_type == "archived-raw-search-results":
            return path.is_dir()
        elif self.base_type == "archived-parsed-search-results":
            return path.is_file() and path.name.endswith(".jsonl.gz")
        else:
            raise ValueError(f"Unknown base type: {self.base_type}")

    def _indexable_paths(self) -> Iterator[Path]:
        base_path = self.base_path
        if not base_path.exists():
            return
        for service_path in base_path.iterdir():
            if not service_path.is_dir():
                continue
            if service_path.name.startswith("."):
                continue
            for pattern_path in service_path.iterdir():
                if (not pattern_path.is_dir() or
                        pattern_path.name.startswith(".")):
                    continue
                for path in pattern_path.iterdir():
                    if self._is_indexable_path(path):
                        yield path

    def _index_path(self, path: Path) -> Path:
        if self.base_type == "archived-urls":
            return path.with_name(
                f"{path.name.removesuffix('.jsonl.gz')}.index"
            )
        elif self.base_type == "archived-query-urls":
            return path.with_name(
                f"{path.name.removesuffix('.jsonl.gz')}.index"
            )
        elif self.base_type == "archived-raw-serps":
            return path.with_name(
                f"{path.name}.index"
            )
        elif self.base_type == "archived-parsed-serps":
            return path.with_name(
                f"{path.name.removesuffix('.jsonl.gz')}.index"
            )
        elif self.base_type == "archived-search-result-snippets":
            return path.with_name(
                f"{path.name.removesuffix('.jsonl.gz')}.snippets.index"
            )
        elif self.base_type == "archived-raw-search-results":
            return path.with_name(
                f"{path.name}.index"
            )
        elif self.base_type == "archived-parsed-search-results":
            return path.with_name(
                f"{path.name.removesuffix('.jsonl.gz')}.index"
            )
        else:
            raise ValueError(f"Unknown base type: {self.base_type}")

    def _index_jsonl(self, path: Path) -> None:
        if not path.exists():
            return
        index_path = self._index_path(path)
        if index_path.exists():
            if (index_path.stat().st_size == 0 or
                    index_path.stat().st_mtime < path.stat().st_mtime):
                # Remove empty or stale index.
                index_path.unlink()
            else:
                # Index is up-to-date.
                return

        offset = 0
        index: list[tuple[str, str, str]] = []
        with (GzipFile(path, mode="rb") as gzip_file,
              text_io_wrapper(gzip_file) as file):
            for line in file:
                try:
                    record = loads(line)
                except JSONDecodeError:
                    LOGGER.error(f"Could not index {line} at {path}.")
                    return
                record_id = uuid5(
                    NAMESPACE_URL,
                    f"{record['timestamp']}:{record['url']}",
                )
                index.append((
                    str(record_id),
                    str(path.relative_to(self.data_directory)),
                    str(offset),
                ))
                offset = gzip_file.tell()

        with index_path.open("wt") as index_file:
            index_writer = writer(index_file)
            index_writer.writerows(index)

    def _index_warc(self, dir_path: Path) -> None:
        if not dir_path.exists():
            return
        index_path = self._index_path(dir_path)
        if index_path.exists():
            if (index_path.stat().st_size == 0 or
                    index_path.stat().st_mtime < dir_path.stat().st_mtime):
                # Remove empty or stale index.
                index_path.unlink()
            else:
                # Index is up-to-date.
                return

        index: list[tuple[str, str, str]] = []
        for path in dir_path.iterdir():
            if path.name.startswith("."):
                continue
            with path.open("rb") as file:
                records = ArchiveIterator(file, no_record_parse=True)
                record: ArcWarcRecord
                for record in records:
                    if record.rec_type != "response":
                        continue
                    offset = record.raw_stream.tell()
                    try:
                        record_url = loads(
                            record.rec_headers.get_header("Archived-URL"))
                    except JSONDecodeError:
                        LOGGER.error(
                            f"Could not index "
                            f"{record.rec_headers.get_header('Archived-URL')} "
                            f"at {path}."
                        )
                        return
                    record_id = uuid5(
                        NAMESPACE_URL,
                        f"{record_url['timestamp']}:{record_url['url']}",
                    )
                    index.append((
                        str(record_id),
                        str(path.relative_to(self.data_directory)),
                        str(offset),
                    ))

        with index_path.open("wt") as index_file:
            index_writer = writer(index_file)
            index_writer.writerows(index)

    def _index_jsonl_snippets(self, path: Path) -> None:
        if not path.exists():
            return
        index_path = self._index_path(path)
        if index_path.exists():
            if (index_path.stat().st_size == 0 or
                    index_path.stat().st_mtime < path.stat().st_mtime):
                # Remove empty or stale index.
                index_path.unlink()
            else:
                # Index is up-to-date.
                return

        offset = 0
        index: list[tuple[str, str, str, str]] = []
        with (GzipFile(path, mode="rb") as gzip_file,
              text_io_wrapper(gzip_file) as file):
            for line in file:
                try:
                    record = loads(line)
                except JSONDecodeError:
                    LOGGER.error(f"Could not index {line} at {path}.")
                    return
                for snippet_index, snippet in enumerate(record["results"]):
                    record_id = uuid5(
                        NAMESPACE_URL,
                        f"{record['timestamp']}:{snippet['url']}",
                    )
                    index.append((
                        str(record_id),
                        str(path.relative_to(self.data_directory)),
                        str(offset),
                        str(snippet_index)
                    ))
                offset = gzip_file.tell()

        with index_path.open("wt") as index_file:
            index_writer = writer(index_file)
            index_writer.writerows(index)

    def _index(self, path: Path) -> None:
        if self.base_type == "archived-urls":
            self._index_jsonl(path)
        elif self.base_type == "archived-query-urls":
            self._index_jsonl(path)
        elif self.base_type == "archived-raw-serps":
            self._index_warc(path)
        elif self.base_type == "archived-parsed-serps":
            self._index_jsonl(path)
        elif self.base_type == "archived-search-result-snippets":
            self._index_jsonl_snippets(path)
        elif self.base_type == "archived-raw-search-results":
            self._index_warc(path)
        elif self.base_type == "archived-parsed-search-results":
            self._index_jsonl(path)
        else:
            raise ValueError(f"Unknown base type: {self.base_type}")

    @cached_property
    def path(self) -> Path:
        if self.base_type == "archived-urls":
            return self.base_path / ".index"
        elif self.base_type == "archived-query-urls":
            return self.base_path / ".index"
        elif self.base_type == "archived-raw-serps":
            return self.base_path / ".index"
        elif self.base_type == "archived-parsed-serps":
            return self.base_path / ".index"
        elif self.base_type == "archived-search-result-snippets":
            return self.base_path / ".snippets.index"
        elif self.base_type == "archived-raw-search-results":
            return self.base_path / ".index"
        elif self.base_type == "archived-parsed-search-results":
            return self.base_path / ".index"
        else:
            raise ValueError(f"Unknown base type: {self.base_type}")

    @cached_property
    def shelf_path(self) -> Path:
        return self.path.with_name(f"{self.path.name}.shelf")

    def index(self) -> None:
        # Index each path individually.
        indexable_paths: Iterable[Path]
        # noinspection PyTypeChecker
        indexable_paths = tqdm(
            self._indexable_paths(),
            total=sum(1 for _ in self._indexable_paths()),
            desc="Index paths",
            unit="path",
        )
        indexed_paths_list: list[Path] = []
        for indexable_path in indexable_paths:
            self._index(indexable_path)
            indexed_paths_list.append(indexable_path)
        indexed_paths: Iterable[Path] = indexed_paths_list

        # Merge all indexes into a single index.
        path = self.path
        num_lines = 0
        with path.open("wb") as aggregated_index_file:
            # noinspection PyTypeChecker
            indexed_paths = tqdm(
                indexed_paths,
                desc="Merge indices",
                unit="path",
            )
            for indexed_path in indexed_paths:
                index_path = self._index_path(indexed_path)
                if not index_path.exists():
                    continue
                with index_path.open("rb") as index_file:
                    copyfileobj(index_file, aggregated_index_file)
                with index_path.open("rb") as index_file:
                    num_lines += count_lines(index_file)


_CorpusLocationType = TypeVar(
    "_CorpusLocationType",
    CorpusJsonlLocation, CorpusJsonlSnippetLocation, CorpusWarcLocation
)
_RecordType = TypeVar("_RecordType", bound=DataClassJsonMixin)


@dataclass(frozen=True)
class LocatedRecord(Generic[_CorpusLocationType, _RecordType]):
    location: _CorpusLocationType
    record: _RecordType


_T = TypeVar("_T")


@dataclass(frozen=True)
class _Index(
    Generic[_CorpusLocationType, _RecordType],
    ContextManager,
    ABC,
):
    data_directory: Path
    focused: bool

    @property
    @abstractmethod
    def base_type(self) -> str:
        pass

    @cached_property
    def _meta_index(self) -> _MetaIndex:
        return _MetaIndex(self.base_type, self.data_directory, self.focused)

    @abstractmethod
    def _to_corpus_location(self, csv_line: list) -> _CorpusLocationType:
        pass

    @cached_property
    def _index_shelf_path(self) -> Path:
        index_path = self.data_directory
        if self.focused:
            index_path /= "focused"
        if self.base_type == "archived-search-result-snippets":
            return index_path / "archived-parsed-serps" / \
                ".snippets.index.shelf"
        else:
            return index_path / self.base_type / ".index.shelf"

    @cached_property
    def _index_shelve(self) -> Shelf:
        return shelf_open(str(self._index_shelf_path), "r")  # nosec: B301

    def index(self) -> None:
        self._meta_index.index()

    @abstractmethod
    def _read_record(self, location: _CorpusLocationType) -> _RecordType:
        pass

    def __getitem__(
            self,
            item: UUID
    ) -> LocatedRecord[_CorpusLocationType, _RecordType]:
        csv_line = self._index_shelve[str(item)].split(",")
        location = self._to_corpus_location(csv_line)
        record = self._read_record(location)
        return LocatedRecord(location, record)

    def get(
            self,
            item: UUID
    ) -> LocatedRecord[_CorpusLocationType, _RecordType] | None:
        if str(item) not in self._index_shelve:
            return None
        return self[item]

    def __iter__(self) -> Iterator[UUID]:
        for uuid in self._index_shelve:
            yield UUID(uuid)

    def __exit__(self, *args) -> None:
        self._index_shelve.close()


@dataclass(frozen=True)
class _JsonLineIndex(_Index[CorpusJsonlLocation, _RecordType]):
    @property
    @abstractmethod
    def record_type(self) -> Type[_RecordType]:
        pass

    @cached_property
    def _schema(self) -> Schema:
        return self.record_type.schema()

    def _to_corpus_location(self, csv_line: list) -> CorpusJsonlLocation:
        return CorpusJsonlLocation(
            relative_path=Path(csv_line[1]),
            byte_offset=int(csv_line[2]),
        )

    def _read_record(self, location: CorpusJsonlLocation) -> _RecordType:
        path = self.data_directory / location.relative_path
        with GzipFile(path, "rb") as gzip_file:
            gzip_file.seek(location.byte_offset)
            with text_io_wrapper(gzip_file) as text_file:
                line = text_file.readline()
                return self._schema.loads(line)


@dataclass(frozen=True)
class _WarcIndex(_Index[CorpusWarcLocation, _RecordType]):

    def _to_corpus_location(self, csv_line: list) -> CorpusWarcLocation:
        return CorpusWarcLocation(
            relative_path=Path(csv_line[1]),
            byte_offset=int(csv_line[2]),
        )

    def _read_record(self, location: CorpusWarcLocation) -> _RecordType:
        path = self.data_directory / location.relative_path
        with path.open("rb") as file:
            file.seek(location.byte_offset)
            record: ArcWarcRecord = next(ArchiveIterator(file))
            return self._read_warc_record(record)

    @abstractmethod
    def _read_warc_record(self, record: ArcWarcRecord) -> _RecordType:
        pass


@dataclass(frozen=True)
class ArchivedUrlIndex(_JsonLineIndex[ArchivedUrl]):
    base_type = final("archived-urls")
    record_type = ArchivedUrl

    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False


@dataclass(frozen=True)
class ArchivedQueryUrlIndex(_JsonLineIndex[ArchivedQueryUrl]):
    base_type = "archived-query-urls"
    record_type = ArchivedQueryUrl

    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False


@dataclass(frozen=True)
class ArchivedRawSerpIndex(_WarcIndex[ArchivedRawSerp]):
    base_type = "archived-raw-serps"
    schema = ArchivedQueryUrl.schema()

    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False

    def _read_warc_record(self, record: ArcWarcRecord) -> ArchivedRawSerp:
        header = record.rec_headers.get_header("Archived-URL")
        archived_url = self.schema.loads(
            record.rec_headers.get_header("Archived-URL"))
        if isinstance(archived_url, list):
            raise ValueError(f"Expected one URL in the header: {header}")
        content_type = record.http_headers.get_header("Content-Type")
        if content_type is None:
            content_type = "utf8"
        return ArchivedRawSerp(
            url=archived_url.url,
            timestamp=archived_url.timestamp,
            query=archived_url.query,
            page=archived_url.page,
            offset=archived_url.offset,
            content=record.content_stream().read(),
            encoding=content_type,
        )


@dataclass(frozen=True)
class ArchivedParsedSerpIndex(_JsonLineIndex[ArchivedParsedSerp]):
    base_type = "archived-parsed-serps"
    record_type = ArchivedParsedSerp

    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False


@dataclass(frozen=True)
class ArchivedSearchResultSnippetIndex(
    _Index[CorpusJsonlSnippetLocation, ArchivedSearchResultSnippet]
):
    base_type = "archived-search-result-snippets"
    schema = ArchivedParsedSerp.schema()

    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False

    def _to_corpus_location(
            self,
            csv_line: list
    ) -> CorpusJsonlSnippetLocation:
        return CorpusJsonlSnippetLocation(
            relative_path=Path(csv_line[1]),
            byte_offset=int(csv_line[2]),
            index=int(csv_line[3]),
        )

    def _read_record(
            self,
            location: CorpusJsonlSnippetLocation
    ) -> ArchivedSearchResultSnippet:
        path = self.data_directory / location.relative_path
        with GzipFile(path, "rb") as gzip_file:
            gzip_file.seek(location.byte_offset)
            with text_io_wrapper(gzip_file) as text_file:
                line = text_file.readline()
                record = self.schema.loads(line)
                if isinstance(record, list):
                    raise ValueError(f"Expected one result per line: {line}")
                return record.results[location.index]


@dataclass(frozen=True)
class ArchivedRawSearchResultIndex(_WarcIndex[ArchivedRawSearchResult]):
    base_type = "archived-raw-search-results"
    schema = ArchivedSearchResultSnippet.schema()

    data_directory: Path = DATA_DIRECTORY_PATH
    focused: bool = False

    def _read_warc_record(
            self,
            record: ArcWarcRecord,
    ) -> ArchivedRawSearchResult:
        header = record.rec_headers.get_header("Archived-URL")
        archived_url = self.schema.loads(header)
        if isinstance(archived_url, list):
            raise ValueError(f"Expected one URL in the header: {header}")
        content_type = record.http_headers.get_header("Content-Type")
        if content_type is None:
            content_type = "utf8"
        return ArchivedRawSearchResult(
            url=archived_url.url,
            timestamp=archived_url.timestamp,
            rank=archived_url.rank,
            title=archived_url.title,
            snippet=archived_url.snippet,
            content=record.content_stream().read(),
            encoding=content_type,
        )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_amazon_under_armour_socks_1553107494():
    verify_serp_parsing(
        "https://web.archive.org/web/20190320184454id_/https://www.amazon.ca/s?k=Under+armour+socks&rh=n%253A2242989011%252Cn%253A2439137011&s=price-asc-rank&dc&qid=1553111078&rnid=5264023011&ref=sr_nr_n_14",
        "amazon",
    )


def test_parse_query_amazon_joseph_menn_1605789427():
    verify_serp_parsing(
        "https://web.archive.org/web/20201119123707id_/https://www.amazon.com/s?k=Joseph+Menn",
        "amazon",
    )


def test_parse_query_amazon_lovense_1618431049():
    verify_serp_parsing(
        "https://web.archive.org/web/20210414201049id_/https://www.amazon.com/s?k=LOVENSE&i=hpc&ref=nb_sb_noss_2",
        "amazon",
    )


def test_parse_query_amazon_notebook_dell_1613232661():
    verify_serp_parsing(
        "https://web.archive.org/web/20210213161101id_/https://www.amazon.com.br/s?k=notebook+dell&__mk_pt_BR=%C3%85M%C3%85%C5%BD%C3%95%C3%91&linkCode=sl2&tag=katlivros-20&linkId=e69cdc65575812a23eb13b5de77ffae8&language=pt_BR&ref_=as_li_ss_tl",
        "amazon",
    )


def test_parse_query_amazon_deng_shi_jia_meng_www_baidu_com_aaaa_oq1_2018nian_8yue_19ri_19shi_32fen_18miao_1564357198():
    verify_serp_parsing(
        "https://web.archive.org/web/20190728233958id_/https://www.amazon.cn/s?k=%E7%81%AF%E9%A5%B0%E5%8A%A0%E7%9B%9F+www.baidu.com%2Faaaa+oq1+2018%E5%B9%B48%E6%9C%8819%E6%97%A519%E6%97%B632%E5%88%8618%E7%A7%92",
        "amazon",
    )


def test_parse_query_amazon_tazas_te_transparente_1613035337():
    verify_serp_parsing(
        "https://web.archive.org/web/20210211092217id_/https://www.amazon.es/s?k=tazas+te+transparente&__mk_es_ES=%C3%85M%C3%85%C5%BD%C3%95%C3%91&crid=12A9DD0FYHVU2&sprefix=tazas+te+tra,aps,244&linkCode=sl2&tag=loxlo3-21&linkId=3e3ee226a50fb1466af6122597a0551c&language=es_ES&ref_=as_li_ss_tl",
        "amazon",
    )


def test_parse_query_amazon_pillow_fsa_1615964514():
    verify_serp_parsing(
        "https://web.archive.org/web/20210317070154id_/https://www.amazon.com/s?i=aps&k=pillow+fsa&ref=nb_sb_noss_1&url=search-alias%3Daps&pldnSite=1",
        "amazon",
    )


def test_parse_query_amazon_yangumagazin_yanmagasado_1625727694():
    verify_serp_parsing(
        "https://web.archive.org/web/20210708070134id_/https://www.amazon.co.jp/s?k=%22%E3%83%A4%E3%83%B3%E3%82%B0%E3%83%9E%E3%82%AC%E3%82%B8%E3%83%B3%22%7C%22%E3%83%A4%E3%83%B3%E3%83%9E%E3%82%AC%E3%82%B5%E3%83%BC%E3%83%89%22&x=0&y=0&linkCode=ll2&tag=rapfoodnewe0c-22&linkId=76c8bdecef23a7536b9c5ec384e267da&language=ja_JP&ref_=as_li_ss_tl",
        "amazon",
    )


def test_parse_query_amazon_monitor_1625069899():
    verify_serp_parsing(
        "https://web.archive.org/web/20210630161819id_/https://www.amazon.de/s?k=monitor&rh=p_n_condition-type%3A776950031&dc&__mk_de_DE=AMAZON&qid=1625076577&rnid=776942031",
        "amazon",
    )


def test_parse_query_amazon_gaming_monitor_1639065822():
    verify_serp_parsing(
        "https://web.archive.org/web/20211209160342id_/https://www.amazon.ca/s?k=Gaming+Monitor&_encoding=UTF8&camp=15121&creative=330641&linkCode=ur2&linkId=769e0014053b086f96c5e9f93948f632&tag=liquipedia07-20",
        "amazon",
    )


def test_parse_query_amazon_the_risks_of_prescription_drugs_1657820625():
    verify_serp_parsing(
        "https://web.archive.org/web/20220714174345id_/https://www.amazon.com/s?k=the+risks+of+prescription+drugs&x=15&y=19&ref=nb_sb_noss",
        "amazon",
    )


def test_parse_query_amazon_eksa_1662208504():
    verify_serp_parsing(
        "https://web.archive.org/web/20220903123504id_/https://www.amazon.in/s?k=eksa&ref=nb_sb_noss&tag=prosenjit0c-21",
        "amazon",
    )


def test_parse_query_amazon_qian_lian_mo_hua_orizinarusaundotoratsuku_1657685343():
    verify_serp_parsing(
        "https://web.archive.org/web/20220713040903id_/https://www.amazon.co.jp/s?k=%E5%8D%83%E6%81%8B%EF%BC%8A%E4%B8%87%E8%8A%B1+%E3%82%AA%E3%83%AA%E3%82%B8%E3%83%8A%E3%83%AB%E3%82%B5%E3%82%A6%E3%83%B3%E3%83%89%E3%83%88%E3%83%A9%E3%83%83%E3%82%AF&__mk_zh_CN=%E4%BA%9A%E9%A9%AC%E9%80%8A%E7%BD%91%E7%AB%99&crid=BFBUJUC89AS1&sprefix=%E5%8D%83%E6%81%8B+%E4%B8%87%E8%8A%B1+original+soundtrack%2Caps%2C546&ref=nb_sb_noss",
        "amazon",
    )


def test_parse_query_amazon_zao_dian_jia_meng_pin_pai_www_baidu_com_aaaa_zv9_2018nian_8yue_21ri_17shi_49fen_4miao_1564430721():
    verify_serp_parsing(
        "https://web.archive.org/web/20190729200521id_/https://www.amazon.cn/s?k=%E6%97%A9%E7%82%B9%E5%8A%A0%E7%9B%9F%E5%93%81%E7%89%8C+www.baidu.com%2Faaaa+zv9+2018%E5%B9%B48%E6%9C%8821%E6%97%A517%E6%97%B649%E5%88%864%E7%A7%92",
        "amazon",
    )


def test_parse_query_amazon_gym_mat_tile_1634791656():
    verify_serp_parsing(
        "https://web.archive.org/web/20211021044736id_/https://www.amazon.com/s?k=gym+mat+tile&ref=nb_sb_noss_1",
        "amazon",
    )


def test_parse_query_amazon_guang_huaiba_pawameta_1632443655():
    verify_serp_parsing(
        "https://web.archive.org/web/20210924003415id_/https://www.amazon.co.jp/s?k=%E5%85%89%E3%83%95%E3%82%A1%E3%82%A4%E3%83%90%E3%83%BC+%E3%83%91%E3%83%AF%E3%83%BC%E3%83%A1%E3%83%BC%E3%82%BF%E3%83%BC&ref=nb_sb_noss_2",
        "amazon",
    )


def test_parse_query_amazon_omega_labyrinth_z_1616510370():
    verify_serp_parsing(
        "https://web.archive.org/web/20210323143930id_/https://www.amazon.com/s?k=omega+labyrinth+z",
        "amazon",
    )


def test_parse_query_amazon_solefit_1663015835():
    verify_serp_parsing(
        "https://web.archive.org/web/20220912205035id_/https://www.amazon.in/s?k=Solefit&bbn=1968024031&rh=n:1571271031,n:1968024031,n:1968120031,p_36:12700-&dc=&qid=1626028382&rnid=4595083031&linkCode=sl2&tag=couponguruj0b-21&linkId=16c64e11c927eb2d1115416374b80d53&language=en_IN&ref_=as_li_ss_tl",
        "amazon",
    )


def test_parse_query_amazon_the_montells_you_can_t_make_me_1617199271():
    verify_serp_parsing(
        "https://web.archive.org/web/20210331140111id_/https://music.amazon.com/search/The+Montells+You+Can't+Make+Me",
        "amazon",
    )


def test_parse_query_amazon_dnd_flash_mosquito_1633240377():
    verify_serp_parsing(
        "https://web.archive.org/web/20211003055257id_/https://www.amazon.in/s?k=DND+Flash+Mosquito&i=lawngarden&rh=p_6:AT95IG9ONZD7S&s=price-asc-rank&dc=&linkCode=sl2&-21=&language=en_IN&ref_=as_li_ss_tl&tag=dealsnalert-21",
        "amazon",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_cnn_kindly_check_1642508434():
    verify_serp_parsing(
        "https://web.archive.org/web/20220118122034id_/https://edition.cnn.com/search?q=kindly%20check",
        "cnn",
    )


def test_parse_query_cnn_achievable_possibility_1642416655():
    verify_serp_parsing(
        "https://web.archive.org/web/20220117105055id_/https://edition.cnn.com/search?q=achievable%20possibility",
        "cnn",
    )


def test_parse_query_cnn_apple_watch_2016_1662575890():
    verify_serp_parsing(
        "https://web.archive.org/web/20220907183810id_/https://www.cnn.com/search?q=apple%20watch%202016",
        "cnn",
    )


def test_parse_query_cnn_march_4_1633908556():
    verify_serp_parsing(
        "https://web.archive.org/web/20211010232916id_/https://www.cnn.com/search?q=March+4",
        "cnn",
    )


def test_parse_query_cnn_dxa_1634498874():
    verify_serp_parsing(
        "https://web.archive.org/web/20211017192754id_/https://www.cnn.com/search?q=dxa",
        "cnn",
    )


def test_parse_query_cnn_trigger_1638271567():
    verify_serp_parsing(
        "https://web.archive.org/web/20211130112607id_/https://edition.cnn.com/search?q=trigger",
        "cnn",
    )


def test_parse_query_cnn_wright_1616229078():
    verify_serp_parsing(
        "https://web.archive.org/web/20210320083118id_/https://www.cnn.com/search?size=10&q=Wright",
        "cnn",
    )


def test_parse_query_cnn_skilled_1643388398():
    verify_serp_parsing(
        "https://web.archive.org/web/20220128164638id_/https://edition.cnn.com/search?q=skilled",
        "cnn",
    )


def test_parse_query_cnn_biteukoinsaryeomyeon_www_99m_kr_bei_biteukoinsaryelie_biteukoinsaeobbbiteukoinsaeobjadeungrog_biteukoinsayongbangbeob8reduplicate_1647153716():
    verify_serp_parsing(
        "https://web.archive.org/web/20220313064156id_/https://edition.cnn.com/search?q=%EB%B9%84%ED%8A%B8%EC%BD%94%EC%9D%B8%EC%82%AC%EB%A0%A4%EB%A9%B4%28WWW%C2%B899M%C2%B8KR%29%E5%AD%9B%EB%B9%84%ED%8A%B8%EC%BD%94%EC%9D%B8%EC%82%AC%EB%A1%80%E5%8A%BD%EB%B9%84%ED%8A%B8%EC%BD%94%EC%9D%B8%EC%82%AC%EC%97%85%E0%AA%AC%EB%B9%84%ED%8A%B8%EC%BD%94%EC%9D%B8%EC%82%AC%EC%97%85%EC%9E%90%EB%93%B1%EB%A1%9D%E2%97%81%EB%B9%84%ED%8A%B8%EC%BD%94%EC%9D%B8%EC%82%AC%EC%9A%A9%EB%B0%A9%EB%B2%958%E2%83%A3reduplicate/",
        "cnn",
    )


def test_parse_query_cnn_north_dakota_coronavirus_1632793271():
    verify_serp_parsing(
        "https://web.archive.org/web/20210928014111id_/https://www.cnn.com/search?q=North+Dakota+coronavirus",
        "cnn",
    )



from pathlib import Path
from typing import Mapping

from marshmallow import ValidationError
from yaml import safe_load

from archive_query_log.legacy.model import Service


def read_services(
        path: Path, ignore_parsing_errors=True
) -> Mapping[str, Service]:
    with path.open("r") as file:
        services_dict = safe_load(file)
        services = []

        for service_dict in services_dict:
            try:
                service = Service.schema(unknown="exclude").load(service_dict)
                services += [(service.name, service)]
            except ValidationError as e:
                if not ignore_parsing_errors:
                    raise ValueError(
                        f"Could not parse service {service_dict}") from e
        if not ignore_parsing_errors:
            service_names = set()
            for name, service in services:
                if name in service_names:
                    raise ValueError(f"Duplicate service name {name}")
                service_names.add(name)
        return {
            name: service
            for name, service in services
        }



from typing import Literal, Type, TypeVar, Iterable
from warnings import warn

from cssselect import GenericTranslator
from cssselect.parser import parse as cssselect_parse
from lxml.etree import parse as etree_parse, XMLParser, HTMLParser  # nosec: B410
# noinspection PyProtectedMember
from lxml.etree import _ElementTree, _Element  # nosec: B410
from warcio.recordloader import ArcWarcRecord

XmlParserType = Literal[
    "xml",
    "html",
]


def parse_xml_tree(record: ArcWarcRecord) -> _ElementTree | None:
    mime_type: str | None = record.http_headers.get_header("Content-Type")
    if mime_type is None:
        warn(UserWarning("No MIME type given."))
        return None
    mime_type = mime_type.split(";", maxsplit=1)[0]
    parser: XMLParser | HTMLParser
    if mime_type == "text/xml":
        parser = XMLParser()
    elif mime_type == "text/html":
        parser = HTMLParser()
    else:
        warn(UserWarning(f"Cannot find XML parser for MIME type: {mime_type}"))
        return None
    return etree_parse(  # nosec: B320
        source=record.content_stream(),
        parser=parser,
    )


_T = TypeVar("_T")


def safe_xpath(
        tree: _ElementTree | _Element,
        xpath: str,
        item_type: Type[_T],
) -> list[_T]:
    results = tree.xpath(xpath, smart_strings=False)
    if not isinstance(results, list):
        results = [results]
    if not all(isinstance(result, item_type) for result in results):
        types = ", ".join({str(type(result)) for result in results})
        raise ValueError(
            f"All results of the XPath '{xpath}' results "
            f"must be of type {item_type}, found: {types}")
    return results


_translator = GenericTranslator()


def xpaths_from_css_selector(css_selector: str) -> list[str]:
    if css_selector == ":--self":
        return ["."]
    selectors = cssselect_parse(css_selector)
    return [
        _translator.selector_to_xpath(
            selector,
            prefix="",
            translate_pseudo_elements=True,
        ).replace("/descendant-or-self::*/", "//")
        for selector in selectors
    ]


def merge_xpaths(xpaths: Iterable[str]) -> str:
    return " | ".join(xpaths)


def text_xpath(
        xpath: str,
        attribute: str | None = None,
        text: bool = False,
) -> str:
    if attribute is None and not text:
        raise ValueError("Either an attribute or text=True must be given.")
    if attribute is not None and text:
        raise ValueError(
            "An attribute and text=True are not allowed at the same time.")
    if text:
        return f"{xpath}//text()"
    else:
        return f"{xpath}/@{attribute}"



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_wikimedia_figures_in_theatrical_costumes_claude_gillot_1673_1722_class_photo_description_french_painter_drawer_1632572254():
    verify_serp_parsing(
        "https://web.archive.org/web/20210925121734id_/https://www.wikidata.org/w/index.php?title=Special:Search&search=Figures+in+Theatrical+Costumes+%0A%0A%0AClaude+Gillot%0A%26nbsp%3B%281673%26ndash%3B1722%29%26nbsp%3B%26nbsp%3B%0A%0A%0A%7Cclass%3Dphoto%0A%0A%0ADescription%0AFrench+painter%2C+drawer+and+graphic+artist%0A%0ADate+of+birth%2Fdeath%0A%0A28+April+1673+%2F+27+April+1673%26nbsp%3B%0A4+May+1722%26nbsp%3B%0A%0ALocation+of+birth%2Fdeath%0A%0ALangres%0AParis%0A%0AWork+location%0A%0AParis%26nbsp%3B%0A%0AAuthority+control%0A%0A%2A%3A+Q559969%0A%2AVIAF%3A%26thinsp%3B27335846%0A%2AISNI%3A%26thinsp%3B0000+0001+2277+7433%0A%2AULAN%3A%26thinsp%3B500115197%0A%2ALCCN%3A%26thinsp%3Bnr93039721%0A%2AWGA%3A%26thinsp%3BGILLOT%2C+Claude%0A%2AWorldCat%0A%0A%0A%0A%0A%0A%0A%0A%0A%0AAntoine+Watteau%0A%26nbsp%3B%281684%26ndash%3B1721%29%26nbsp%3B%26nbsp%3B%26nbsp%3B%0A%0A%0A%7Cclass%3Dphoto%0A%0A%0AAlternative+names%0A%0AJean-Antoine+Watteau%0A%0ADescription%0AFrench+painter%2C+graphic+artist%2C+drawer%2C+artist%2C+printmaker+and+draftsperson%0A%0ADate+of+birth%2Fdeath%0A%0A10+October+1684%26nbsp%3B%0A18+July+1721%26nbsp%3B%0A%0ALocation+of+birth%2Fdeath%0A%0AValenciennes%0ANogent-sur-Marne%0A%0AWork+location%0A%0AParis+%28after+1702+%29%2C+Valenciennes+%281709-1710%29%2C+London+%281719-1720%29%0A%0AAuthority+control%0A%0A%2A%3A+Q183221%0A%2AVIAF%3A%26thinsp%3B46882028%0A%2AISNI%3A%26thinsp%3B0000+0001+2131+1877%0A%2AULAN%3A%26thinsp%3B500032644%0A%2ALCCN%3A%26thinsp%3Bn50043693%0A%2ANLA%3A%26thinsp%3B36213816%0A%2AWorldCat%0A%0A%0A%0A%0A%0A+10.45.15+haswbstatement%3AP195%3DQ160236&profile=advanced&fulltext=1&ns0=1",
        "wikimedia",
    )


def test_parse_query_wikimedia_group_portrait_on_doorstep_collins_tudor_washington_1898_1970_photographer_75426_object_number_haswb_1629837781():
    verify_serp_parsing(
        "https://web.archive.org/web/20210824204301id_/https://www.wikidata.org/w/index.php?search=(Group+portrait+on+doorstep)+Collins,+Tudor+Washington,+1898-1970,+photographer+75426+(object+number)+haswbstatement:P195%3DQ758657&title=Special:Search&profile=advanced&fulltext=1&ns0=1",
        "wikimedia",
    )


def test_parse_query_wikimedia_mou_lin_han_site_pku_edu_cn_1576868395():
    verify_serp_parsing(
        "https://web.archive.org/web/20191220185955id_/https://en.wikipedia.org/w/index.php?title=Special%3ASearch&profile=images&search=%E7%89%9F%E6%9E%97%E7%BF%B0+site%3Apku.edu.cn&fulltext=1&ns0=1",
        "wikimedia",
    )


def test_parse_query_wikimedia_nito_1655733503():
    verify_serp_parsing(
        "https://web.archive.org/web/20220620135823id_/https://commons.wikimedia.org/w/index.php?title=Special%3AMediaSearch&search=Nit%C6%A1",
        "wikimedia",
    )


def test_parse_query_wikimedia_prob_1543207125():
    verify_serp_parsing(
        "https://web.archive.org/web/20181126043845id_/https://zh.wikipedia.org/w/index.php?search=prob&title=Special%3A%E6%90%9C%E7%B4%A2&profile=default&fulltext=1",
        "wikimedia",
    )


def test_parse_query_wikimedia_ao_xian_yu_le_du_bo_ping_tai_1430758639():
    verify_serp_parsing(
        "https://web.archive.org/web/20150504165719id_/http://zh.wikipedia.org/w/index.php?search=%E6%BE%B3%E7%BA%BF%E5%A8%B1%E4%B9%90%E8%B5%8C%E5%8D%9A%E5%B9%B3%E5%8F%B0",
        "wikimedia",
    )


def test_parse_query_wikimedia_zao_can_bao_zi_jia_meng_www_baidu_com_aaaa_4ws_2018nian_8yue_4ri_21shi_19fen_39miao_1537939782():
    verify_serp_parsing(
        "https://web.archive.org/web/20180926052942id_/https://zh.wikipedia.org/w/index.php?search=%E6%97%A9%E9%A4%90%E5%8C%85%E5%AD%90%E5%8A%A0%E7%9B%9F%20www.baidu.com/aaaa%204ws%202018%E5%B9%B48%E6%9C%884%E6%97%A521%E6%97%B619%E5%88%8639%E7%A7%92",
        "wikimedia",
    )


def test_parse_query_wikimedia_oxygen_1596274448():
    verify_serp_parsing(
        "https://web.archive.org/web/20200801093408id_/https://commons.wikimedia.org/w/index.php?title=Special:Search&search=Oxygen&ns0=1&ns6=1&ns12=1&ns14=1&ns100=1&ns106=1",
        "wikimedia",
    )


def test_parse_query_wikimedia_recaptchalogo_svg_1509131182():
    verify_serp_parsing(
        "https://web.archive.org/web/20171027190622id_/https://www.wikidata.org/w/index.php?title=Special:Search&profile=images&search=RecaptchaLogo.svg&fulltext=1",
        "wikimedia",
    )


def test_parse_query_wikimedia_la_dian_zi_ji_1517046182():
    verify_serp_parsing(
        "https://web.archive.org/web/20180127094302id_/https://zh.wikipedia.org/w/index.php?title=Special:%E6%90%9C%E7%B4%A2&search=%E6%8B%89%E9%9B%BB%E5%AD%90%E5%9F%BA",
        "wikimedia",
    )



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import cached_property
from gzip import GzipFile
from itertools import chain
from pathlib import Path
from typing import AbstractSet, Sequence, Any, Iterable, Iterator, NamedTuple
from urllib.parse import urlencode, quote

from aiohttp import ClientResponseError
from aiohttp_retry import RetryClient
from asyncio_pool import AioPool
from diskcache import Cache
from marshmallow import Schema
from tqdm.auto import tqdm

from archive_query_log.legacy import CDX_API_URL
from archive_query_log.legacy.model import ArchivedUrl, Service
from archive_query_log.legacy.util.archive_http import archive_http_client
from archive_query_log.legacy.util.text import text_io_wrapper


class UrlMatchScope(Enum):
    EXACT = "exact"
    PREFIX = "prefix"
    HOST = "host"
    DOMAIN = "domain"


class _CdxPage(NamedTuple):
    url: str
    page: int
    path: Path


@dataclass(frozen=True)
class ArchivedUrlsFetcher:
    """
    Fetch archived URLs from the Wayback Machine's CDX API and
    store them in a JSONL file.
    """

    match_scope: UrlMatchScope = UrlMatchScope.EXACT
    include_status_codes: AbstractSet[int] = frozenset({200})
    exclude_status_codes: AbstractSet[int] = frozenset({})
    include_mime_types: AbstractSet[str] = frozenset({"text/html"})
    exclude_mime_types: AbstractSet[str] = frozenset({})
    cdx_api_url: str = CDX_API_URL
    overwrite: bool = False

    @cached_property
    def _base_params(self) -> Sequence[tuple[Any, Any]]:
        params = [
            ("matchType", self.match_scope.value),
            ("fl", "timestamp,original"),
        ]
        if len(self.include_mime_types) > 0:
            pattern = "|".join(self.include_mime_types)
            params.append(("filter", f"mimetype:{pattern}"))
        if len(self.exclude_mime_types) > 0:
            pattern = "|".join(self.exclude_mime_types)
            params.append(("filter", f"mimetype:{pattern}"))
        if len(self.include_status_codes) > 0:
            pattern = "|".join(str(code) for code in self.include_status_codes)
            params.append(("filter", f"statuscode:{pattern}"))
        if len(self.exclude_status_codes) > 0:
            pattern = "|".join(str(code) for code in self.exclude_status_codes)
            params.append(("filter", f"statuscode:{pattern}"))
        return params

    def _params(self, url: str) -> Sequence[tuple[Any, Any]]:
        return [
            ("url", url),
            *self._base_params,
        ]

    async def _num_pages(
            self,
            cache: Cache,
            url: str,
            client: RetryClient,
    ) -> int:
        params = self._params(url)
        params_hash = urlencode(params)
        if params_hash in cache:
            return cache[params_hash]
        num_pages_params = [
            *params,
            ("showNumPages", True),
        ]
        url = f"{self.cdx_api_url}?{urlencode(num_pages_params)}"
        async with client.get(url) as response:
            text = await response.text()
            # noinspection PyBroadException
            try:
                num_pages = int(text)
            except Exception:
                num_pages = 0
        cache[params_hash] = num_pages
        return num_pages

    @staticmethod
    def _parse_response_lines(
            lines: Iterable[str],
            schema: Schema,
    ) -> Iterator[str]:
        for line in lines:
            line = line.strip()
            timestamp_string, url = line.split()
            timestamp = datetime.strptime(
                timestamp_string,
                "%Y%m%d%H%M%S"
            )
            archived_url = ArchivedUrl(
                url=url,
                timestamp=int(timestamp.timestamp()),
            )
            yield schema.dumps(archived_url)

    async def _fetch_page(
            self,
            page: _CdxPage,
            client: RetryClient,
            progress: tqdm | None = None,
    ) -> None:
        if page.path.exists() and not self.overwrite and progress is not None:
            progress.update()
            return
        params = [
            *self._params(page.url),
            ("page", page.page),
        ]
        url = f"{self.cdx_api_url}?{urlencode(params)}"
        try:
            async with client.get(url) as response:
                response.raise_for_status()
                text = await response.text()
                schema = ArchivedUrl.schema()
                lines = self._parse_response_lines(
                    text.splitlines(keepends=False),
                    schema,
                )
                page.path.parent.mkdir(parents=True, exist_ok=True)
                with page.path.open("wb") as file, \
                        GzipFile(fileobj=file, mode="wb") as gzip_file, \
                        text_io_wrapper(gzip_file) as text_file:
                    for line in lines:
                        text_file.write(line)
                        text_file.write("\n")
                return
        except ClientResponseError as e:
            page.path.unlink(missing_ok=True)
            print(
                f"HTTP error {e.status} when fetching {url}. "
                f"Please try again later. Continuing with next URL."
            )
            return None
        except BaseException as e:
            page.path.unlink(missing_ok=True)
            raise e
        finally:
            if progress is not None:
                progress.update()

    async def _service_pages(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None,
            cdx_page: int | None,
            client: RetryClient,
    ) -> Sequence[_CdxPage]:
        """
        List all items that need to be downloaded.
        """
        output_format_path = data_directory / "archived-urls"
        output_format_path.mkdir(parents=True, exist_ok=True)
        if cdx_page is not None:
            if domain is None:
                raise RuntimeError(
                    "Domain must be specified when page is specified.")
            service_path = output_format_path / service.name
            domain_path = service_path / quote(domain, safe="")
            cdx_page_path = domain_path / f"{cdx_page:010}.jsonl.gz"
            return [
                _CdxPage(
                    path=cdx_page_path,
                    url=domain,
                    page=cdx_page,
                )
            ]
        elif domain is not None:

            async def cdx_page_pages(_cdx_page: int) -> Sequence[_CdxPage]:
                return await self._service_pages(
                    data_directory=data_directory,
                    focused=focused,
                    service=service,
                    domain=domain,
                    cdx_page=_cdx_page,
                    client=client,
                )

            with Cache(str(output_format_path / ".pages")) as cache:
                num_cdx_pages = await self._num_pages(
                    cache,
                    domain,
                    client,
                )
            pool = AioPool(size=1)

            if num_cdx_pages <= 0:
                return []

            return list(chain.from_iterable(
                await pool.map(cdx_page_pages, range(num_cdx_pages))
            ))
        else:
            domains = service.domains
            if focused:
                domains = [
                    f"{domain}{url_prefix}"
                    for domain in domains
                    for url_prefix in service.focused_url_prefixes
                ]
            else:
                suffix_free_domains: list[str] = []
                for domain in sorted(domains, key=len):
                    if not any(
                            domain.endswith(suffix)
                            for suffix in suffix_free_domains
                    ):
                        suffix_free_domains.append(domain)
                domains = suffix_free_domains

            if len(domains) == 0:
                return []

            domains = sorted(domains)
            progress = tqdm(
                domains,
                total=len(domains),
                desc="Fetching number of pages",
                unit="domain",
            )

            async def domain_pages(_domain: str) -> Sequence[_CdxPage]:
                res = await self._service_pages(
                    data_directory=data_directory,
                    focused=focused,
                    service=service,
                    domain=_domain,
                    cdx_page=None,
                    client=client,
                )
                progress.update()
                return res

            pool = AioPool(size=100)

            res: Sequence[Sequence[_CdxPage]]
            res = await pool.map(domain_pages, domains)
            for val in res:
                if isinstance(val, Exception):
                    raise val
            res = sorted(res, key=len, reverse=True)
            return list(chain.from_iterable(res))

    async def fetch_service(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None = None,
            cdx_page: int | None = None,
    ):
        async with archive_http_client(limit=5) as client:
            pages = await self._service_pages(
                data_directory=data_directory,
                focused=focused,
                service=service,
                domain=domain,
                cdx_page=cdx_page,
                client=client,
            )

            if len(pages) == 0:
                return

            progress = None
            if len(pages) > 1:
                progress = tqdm(
                    total=len(pages),
                    desc="Fetch archived URLs",
                    unit="page",
                )

            async def fetch_page(page: _CdxPage):
                return await self._fetch_page(
                    page=page,
                    client=client,
                    progress=progress,
                )

            pool = AioPool(size=1000)
            await pool.map(fetch_page, pages)

    async def num_service_pages(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None = None,
            cdx_page: int | None = None,
    ) -> int:
        async with archive_http_client(limit=2) as client:
            pages = await self._service_pages(
                data_directory=data_directory,
                focused=focused,
                service=service,
                domain=domain,
                cdx_page=cdx_page,
                client=client,
            )
            return len(pages)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_bongacams_facial_1578106424():
    verify_serp_parsing(
        "https://web.archive.org/web/20200104025344id_/https://en.bongacams.com/male/tags/facial",
        "bongacams",
    )


def test_parse_query_bongacams_hd_plus_1576443879():
    verify_serp_parsing(
        "https://web.archive.org/web/20191215210439id_/https://bongacams.com/male/tags/hd-plus",
        "bongacams",
    )


def test_parse_query_bongacams_camshow_1653633300():
    verify_serp_parsing(
        "https://web.archive.org/web/20220527063500id_/https://bongacams.com/female/tags/camshow",
        "bongacams",
    )


def test_parse_query_bongacams_stripping_1623091287():
    verify_serp_parsing(
        "https://web.archive.org/web/20210607184127id_/https://bongacams.com/female/tags/stripping?page=3",
        "bongacams",
    )


def test_parse_query_bongacams_teasing_1635992326():
    verify_serp_parsing(
        "https://web.archive.org/web/20211104021846id_/https://bongacams.com/couples/tags/teasing",
        "bongacams",
    )


def test_parse_query_bongacams_asslicking_1623701105():
    verify_serp_parsing(
        "https://web.archive.org/web/20210614200505id_/https://bongacams.com/couples/tags/asslicking",
        "bongacams",
    )


def test_parse_query_bongacams_fucking_1577494848():
    verify_serp_parsing(
        "https://web.archive.org/web/20191228010048id_/https://en.bongacams.com/male/tags/fucking",
        "bongacams",
    )


def test_parse_query_bongacams_cock_sucking_1617355851():
    verify_serp_parsing(
        "https://web.archive.org/web/20210402093051id_/https://bongacams.com/male/tags/cock-sucking?page=2",
        "bongacams",
    )


def test_parse_query_bongacams_massage_1604010948():
    verify_serp_parsing(
        "https://web.archive.org/web/20201029223548id_/https://en.bongacams.com/male/tags/massage",
        "bongacams",
    )


def test_parse_query_bongacams_stripping_1648087966():
    verify_serp_parsing(
        "https://web.archive.org/web/20220324021246id_/https://en.bongacams.com/couples/tags/stripping",
        "bongacams",
    )



from uuid import uuid4

from click import echo, prompt
from elasticsearch_dsl import Search
from elasticsearch_dsl.query import Term
from elasticsearch_dsl.response import Response

from archive_query_log.config import Config
from archive_query_log.orm import Archive
from archive_query_log.utils.time import utc_now


def add_archive(
        config: Config,
        name: str | None,
        description: str | None,
        cdx_api_url: str,
        memento_api_url: str,
        priority: float | None,
        no_merge: bool = False,
        auto_merge: bool = False,
) -> None:
    if priority is not None and priority <= 0:
        raise ValueError("Priority must be strictly positive.")
    config.es.client.indices.refresh(index=config.es.index_archives)
    last_modified = utc_now()
    should_build_sources = True
    existing_archive_search: Search = (
        Archive.search(using=config.es.client, index=config.es.index_archives)
        .query(
            Term(cdx_api_url=cdx_api_url) |
            Term(memento_api_url=memento_api_url)
        )
    )
    existing_archive_response: Response = existing_archive_search.execute()
    if existing_archive_response.hits.total.value > 0:
        if no_merge:
            return
        existing_archive: Archive = existing_archive_response[0]
        archive_id = existing_archive.id
        if auto_merge:
            should_merge = True
        else:
            echo(f"Archive {archive_id} already exists with "
                 f"conflicting API endpoints.")
            add_to_existing = prompt("Merge with existing archive? [y/N]",
                                     type=str, default="n", show_default=False)
            should_merge = add_to_existing.lower() == "y"
        if not should_merge:
            return

        if name is None:
            name = existing_archive.name
        if description is None:
            description = existing_archive.description
        if priority is None:
            priority = existing_archive.priority

        if cdx_api_url == existing_archive.cdx_api_url and \
                memento_api_url == existing_archive.memento_api_url:
            last_modified = existing_archive.last_modified
            should_build_sources = existing_archive.should_build_sources

        if not auto_merge:
            echo(f"Update archive {archive_id}.")
    else:
        archive_id = str(uuid4())
        if not no_merge and not auto_merge:
            echo(f"Add new archive {archive_id}.")

    archive = Archive(
        id=archive_id,
        last_modified=last_modified,
        name=name,
        description=description,
        cdx_api_url=cdx_api_url,
        memento_api_url=memento_api_url,
        priority=priority,
        should_build_sources=should_build_sources,
    )
    archive.save(using=config.es.client, index=config.es.index_archives)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_reddit_teleperformance_1260472045():
    verify_serp_parsing(
        "https://web.archive.org/web/20091210190725id_/http://www.reddit.com:80/search?q=%22teleperformance%22&KeepThis=true&TB_iframe=true&height=550&width=800",
        "reddit",
    )


def test_parse_query_reddit_3ds_1364668924():
    verify_serp_parsing(
        "https://web.archive.org/web/20130330184204id_/http://en.reddit.com/search?q=3ds",
        "reddit",
    )


def test_parse_query_reddit_jailbait_1376244913():
    verify_serp_parsing(
        "https://web.archive.org/web/20130811181513id_/http://www.reddit.com/search?q=jailbait",
        "reddit",
    )


def test_parse_query_reddit_german_nebelmittelwurfanlage_1664557458():
    verify_serp_parsing(
        "https://web.archive.org/web/20220930170418id_/https://www.reddit.com/search/?q=german+nebelmittelwurfanlage&sort=comments&t=all",
        "reddit",
    )


def test_parse_query_reddit_supermoon_1403887175():
    verify_serp_parsing(
        "https://web.archive.org/web/20140627163935id_/http://fr.reddit.com/search?q=supermoon",
        "reddit",
    )


def test_parse_query_reddit_touch_meme_1663170731():
    verify_serp_parsing(
        "https://web.archive.org/web/20220914155211id_/https://www.reddit.com/search/?q=touch%20meme&include_over_18=1&type=link",
        "reddit",
    )


def test_parse_query_reddit_flair_already_posted_https_redd_it_xamfzl_1667030236():
    verify_serp_parsing(
        "https://web.archive.org/web/20221029075716id_/https://old.reddit.com/search/?q=flair%3A%22already+posted+https%3A%2F%2Fredd.it%2Fxamfzl%22&sort=comments&type=comment",
        "reddit",
    )


def test_parse_query_reddit_how_can_i_access_nature_com_articles_1666041985():
    verify_serp_parsing(
        "https://web.archive.org/web/20221017212625id_/https://www.reddit.com/search/?q=How%20Can%20I%20Access%20Nature.com%20Articles%3F&sort=top&type=comment",
        "reddit",
    )


def test_parse_query_reddit_b7_1650853917():
    verify_serp_parsing(
        "https://web.archive.org/web/20220425023157id_/https://www.reddit.com/search/?q=b7",
        "reddit",
    )


def test_parse_query_reddit_typetest_10fastfingers_1663797626():
    verify_serp_parsing(
        "https://web.archive.org/web/20220921220026id_/https://old.reddit.com/search?q=typetest+10fastfingers&restrict_sr=&sort=relevance&t=all",
        "reddit",
    )



from typing import Iterator, TypeVar, Iterable, Any
from warnings import warn

from elasticsearch import NotFoundError
from elasticsearch_dsl import Document, InnerDoc
from elasticsearch_dsl.utils import META_FIELDS

DocumentType = TypeVar("DocumentType", bound=Document)


def safe_iter_scan(it: Iterable[DocumentType]) -> Iterator[DocumentType]:
    try:
        for doc in it:
            yield doc
    except NotFoundError as e:
        if (e.info is not None and isinstance(e.info, dict) and
                "error" in e.info and
                isinstance(e.info["error"], dict) and
                "root_cause" in e.info["error"] and
                isinstance(e.info["error"]["root_cause"], list) and
                len(e.info["error"]["root_cause"]) > 0 and
                isinstance(e.info["error"]["root_cause"][0], dict) and
                "resource.id" in e.info["error"]["root_cause"][0] and
                e.info["error"]["root_cause"][0]["resource.id"] ==
                "search_phase_execution_exception"):
            warn(RuntimeWarning("Scan expired. Stopping iteration."))
            raise StopIteration() from e
        else:
            raise e


def _to_dict_if_needed(value: Any) -> Any:
    if isinstance(value, InnerDoc):
        return value.to_dict()
    return value


def update_action(
        document: Document,
        retry_on_conflict: int | None = 3,
        **fields,
) -> dict:
    action = {
        f"_{key}": document.meta[key]
        for key in META_FIELDS
        if key not in ("score")
        if key in document.meta and document.meta[key] is not None
    }
    action["_op_type"] = "update"
    # Create a partial document by instantiating a new object of the
    # document type and ignoring the meta fields (e.g., the document ID).
    action["doc"] = type(document)(**fields).to_dict(
        include_meta=False,
        skip_empty=True,
    )
    if retry_on_conflict is not None:
        action["_retry_on_conflict"] = retry_on_conflict
    return action



from urllib.parse import parse_qsl, urlsplit, unquote


def parse_url_query_parameter(
        parameter: str, url: str) -> str | None:
    for key, value in parse_qsl(urlsplit(url).query):
        if key == parameter:
            return value
    return None


def parse_url_fragment_parameter(
        parameter: str, url: str) -> str | None:
    for key, value in parse_qsl(urlsplit(url).fragment):
        if key == parameter:
            return value
    return None


def parse_url_path_segment(segment: int, url: str) -> str | None:
    path_segments = urlsplit(url).path.split("/")
    if len(path_segments) <= segment:
        return None
    path_segment = path_segments[segment]
    return unquote(path_segment)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_chefkoch_spaghetti_eis_torte_1342866905():
    verify_serp_parsing(
        "https://web.archive.org/web/20120721103505id_/http://www.chefkoch.de:80/rs/s0/,spaghetti+eis+torte/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_walnussl_dressing_1403236002():
    verify_serp_parsing(
        "https://web.archive.org/web/20140620034642id_/http://www.chefkoch.de:80/rs/s0/,walnuss%F6l+dressing/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eier_spinat_pilze_1200619371():
    verify_serp_parsing(
        "https://web.archive.org/web/20080118012251id_/http://www.chefkoch.de:80/rs/s0/eier+spinat+pilze/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_tomaten_mozarella_reis_1317162887():
    verify_serp_parsing(
        "https://web.archive.org/web/20110927223447id_/http://www.chefkoch.de:80/rs/s0/%2C+tomaten+mozarella+reis/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spargel_m_soe_1280861753():
    verify_serp_parsing(
        "https://web.archive.org/web/20100803185553id_/http://www.chefkoch.de:80/rs/s0/spargel+m.+so%DFe/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spanische_salat_1192556276():
    verify_serp_parsing(
        "https://web.archive.org/web/20071016173756id_/http://www.chefkoch.de:80/rs/s0/spanische+salat/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_thailndisch_1313401080():
    verify_serp_parsing(
        "https://web.archive.org/web/20110815093800id_/http://www.chefkoch.de:80/rs/s0/%2Cthail%E4ndisch/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_hhnchen_warme_1303285108():
    verify_serp_parsing(
        "https://web.archive.org/web/20110420073828id_/http://www.chefkoch.de:80/rs/s0/%2Ch%E4hnchen+warme/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spaghettitorte_vom_blech_1369060560():
    verify_serp_parsing(
        "https://web.archive.org/web/20130520143600id_/http://www.chefkoch.de:80/rs/s0/spaghettitorte+vom+blech/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_sahne_und_himbeeren_1192658021():
    verify_serp_parsing(
        "https://web.archive.org/web/20071017215341id_/http://www.chefkoch.de:80/rs/s0/,sahne+und+himbeeren/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spinat_in_1220295440():
    verify_serp_parsing(
        "https://web.archive.org/web/20080901185720id_/http://www.chefkoch.de:80/rs/s0/spinat+in/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_omas_blechkuchen_1355822404():
    verify_serp_parsing(
        "https://web.archive.org/web/20121218092004id_/http://www.chefkoch.de/rs/s0/%2C+omas+blechkuchen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spargel_geschlt_1278411144():
    verify_serp_parsing(
        "https://web.archive.org/web/20100706101224id_/http://www.chefkoch.de:80/rs/s0/spargel+gesch%E4lt/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierschwammerl_rahmsauce_1449833431():
    verify_serp_parsing(
        "https://web.archive.org/web/20151211113031id_/http://www.chefkoch.de/rs/s0/eierschwammerl+rahmsauce/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_35_backen_1351851992():
    verify_serp_parsing(
        "https://web.archive.org/web/20121102102632id_/http://www.chefkoch.de:80/rs/s0/%2C35+backen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierschwammerl_und_gehacktes_1315110709():
    verify_serp_parsing(
        "https://web.archive.org/web/20110904043149id_/http://www.chefkoch.de/rs/s0/eierschwammerl+und+gehacktes/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_thermomix_gefllten_1390802313():
    verify_serp_parsing(
        "https://web.archive.org/web/20140127055833id_/http://www.chefkoch.de:80/rs/s0/thermomix+gef%FCllten/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierlikr_kirschkuchen_1243942587():
    verify_serp_parsing(
        "https://web.archive.org/web/20090602113627id_/http://www.chefkoch.de:80/rs/s0/eierlik%F6r+kirschkuchen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_syrische_pasta_1466000049():
    verify_serp_parsing(
        "https://web.archive.org/web/20160615141409id_/http://www.chefkoch.de:80/rs/s0/syrische+pasta/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_speckbohnen_mit_kartoffeln_1352170361():
    verify_serp_parsing(
        "https://web.archive.org/web/20121106025241id_/http://www.chefkoch.de:80/rs/s0/,+speckbohnen+mit+kartoffeln/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spargel_wei_1194284442():
    verify_serp_parsing(
        "https://web.archive.org/web/20071105174042id_/http://www.chefkoch.de:80/rs/s0/spargel%2Cwei%DF/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_tempura_1213419780():
    verify_serp_parsing(
        "https://web.archive.org/web/20080614050300id_/http://www.chefkoch.de:80/rs/s0/tempura%2A/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_suppen_oder_eintpfe_1256106491():
    verify_serp_parsing(
        "https://web.archive.org/web/20091021062811id_/http://www.chefkoch.de:80/rs/s0/suppen+oder+eint%F6pfe/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eifreier_hackbraten_1356156512():
    verify_serp_parsing(
        "https://web.archive.org/web/20121222060832id_/http://www.chefkoch.de/rs/s0/eifreier+hackbraten/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_kseschnitzel_kse_1353053392():
    verify_serp_parsing(
        "https://web.archive.org/web/20121116080952id_/http://www.chefkoch.de:80/rs/s0/%2Ck%E4seschnitzel+k%E4se/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierlikr_mit_kirsch_1375310772():
    verify_serp_parsing(
        "https://web.archive.org/web/20130731224612id_/http://www.chefkoch.de:80/rs/s0/eierlik%F6r+mit+kirsch/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_thai_curry_1246123954():
    verify_serp_parsing(
        "https://web.archive.org/web/20090627173234id_/http://www.chefkoch.de:80/rs/s0/thai-+curry/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierflaumsuppe_1249321209():
    verify_serp_parsing(
        "https://web.archive.org/web/20090803174009id_/http://www.chefkoch.de:80/rs/s0/eierflaumsuppe/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_tatar_steak_1266004316():
    verify_serp_parsing(
        "https://web.archive.org/web/20100212195156id_/http://www.chefkoch.de:80/rs/s0/tatar+steak/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_nudelsalat_mit_ei_1330099391():
    verify_serp_parsing(
        "https://web.archive.org/web/20120224160311id_/http://www.chefkoch.de:80/rs/s0/%2Cnudelsalat+mit+ei/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spaghettie_mit_garnelen_1206750318():
    verify_serp_parsing(
        "https://web.archive.org/web/20080329002518id_/http://www.chefkoch.de:80/rs/s0/spaghettie+mit+garnelen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_speisestrke_sahne_1232351646():
    verify_serp_parsing(
        "https://web.archive.org/web/20090119075406id_/http://www.chefkoch.de:80/rs/s0/speisest%E4rke+sahne/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierlikr_nachspeise_1255758312():
    verify_serp_parsing(
        "https://web.archive.org/web/20091017054512id_/http://www.chefkoch.de:80/rs/s0/eierlik%F6r+nachspeise/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_roulade_braten_1355445937():
    verify_serp_parsing(
        "https://web.archive.org/web/20121214004537id_/http://www.chefkoch.de:80/rs/s0/%2Croulade+braten/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spaghetti_hhnchen_1192740442():
    verify_serp_parsing(
        "https://web.archive.org/web/20071018204722id_/http://www.chefkoch.de:80/rs/s0/spaghetti+h%E4hnchen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_kohlroulade_gemse_1328779105():
    verify_serp_parsing(
        "https://web.archive.org/web/20120209091825id_/http://www.chefkoch.de/rs/s0/%2Ckohlroulade+gem%FCse/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_sushi_mit_normalem_reis_1470947014():
    verify_serp_parsing(
        "https://web.archive.org/web/20160811202334id_/http://www.chefkoch.de:80/rs/s0/sushi+mit+normalem+reis/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_tabouleh_rosinen_1382360347():
    verify_serp_parsing(
        "https://web.archive.org/web/20131021125907id_/http://www.chefkoch.de:80/rs/s0/tabouleh+-rosinen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_kirsch_1305195838():
    verify_serp_parsing(
        "https://web.archive.org/web/20110512102358id_/http://www.chefkoch.de:80/rs/s0/%2C+kirsch/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_speck_zucchini_eier_sahne_1222747904():
    verify_serp_parsing(
        "https://web.archive.org/web/20080930041144id_/http://www.chefkoch.de:80/rs/s0/speck+zucchini+eier+sahne/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spaghettoni_1206569039():
    verify_serp_parsing(
        "https://web.archive.org/web/20080326220359id_/http://www.chefkoch.de:80/rs/s0/spaghettoni/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spaghetti_mit_gemsesauce_1190132272():
    verify_serp_parsing(
        "https://web.archive.org/web/20070918161752id_/http://www.chefkoch.de:80/rs/s0/spaghetti+mit+gem%FCsesauce/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_spargel_risotto_mit_garnelen_1471699836():
    verify_serp_parsing(
        "https://web.archive.org/web/20160820133036id_/http://www.chefkoch.de/rs/s0/spargel+risotto+mit+garnelen/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_tupperware_hefe_1356254719():
    verify_serp_parsing(
        "https://web.archive.org/web/20121223092519id_/http://www.chefkoch.de/rs/s0/%2C+tupperware+hefe/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eier_milch_kse_kartoffeln_1318228714():
    verify_serp_parsing(
        "https://web.archive.org/web/20111010063834id_/http://www.chefkoch.de:80/rs/s0/eier%2C+milch%2C+k%E4se%2C+kartoffeln/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_speck_n_dicken_1370169101():
    verify_serp_parsing(
        "https://web.archive.org/web/20130602103141id_/http://www.chefkoch.de/rs/s0/speck+n+dicken/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eier_parmesan_spinat_sahne_1206623203():
    verify_serp_parsing(
        "https://web.archive.org/web/20080327130643id_/http://www.chefkoch.de:80/rs/s0/eier+parmesan+spinat+sahne/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_eierlikr_gugelhupf_1203899612():
    verify_serp_parsing(
        "https://web.archive.org/web/20080225003332id_/http://www.chefkoch.de:80/rs/s0/eierlik%F6r+gugelhupf%2C/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_thai_curry_einfach_1420408232():
    verify_serp_parsing(
        "https://web.archive.org/web/20150104215032id_/http://www.chefkoch.de:80/rs/s0/thai+curry+einfach/Rezepte.html",
        "chefkoch",
    )


def test_parse_query_chefkoch_streuselkuchen_aus_mrbeteig_1299585024():
    verify_serp_parsing(
        "https://web.archive.org/web/20110308115024id_/http://www.chefkoch.de:80/rs/s0/%2Cstreuselkuchen+aus+m%FCrbeteig/Rezepte.html",
        "chefkoch",
    )



from contextlib import contextmanager
from typing import Iterator

from warc_s3 import WarcS3Store, WarcS3Location
from warcio.recordloader import ArcWarcRecord

from archive_query_log.orm import WarcLocation


@contextmanager
def open_warc(
        warc_store: WarcS3Store,
        warc_location: WarcLocation,
) -> Iterator[ArcWarcRecord]:
    with warc_store.read(WarcS3Location(
            key=warc_location.file,
            offset=warc_location.offset,
            length=warc_location.length,
    )) as record:
        yield record



from dataclasses import dataclass
from functools import cached_property
from html.parser import HTMLParser
from typing import List, Union, Optional


class Highlight(str):
    depth: int

    def __new__(cls, value: str, depth: int):
        result = super().__new__(cls, value)
        result.depth = depth
        return result


@dataclass(frozen=True, unsafe_hash=True)
class HighlightedText(str):
    """
    Text with highlighting using``<em>`` tags.
    Other tags are not allowed.
    """
    html: str

    @property
    def text(self):
        return "".join(self.sequence)

    def __str__(self) -> str:
        return self.text

    @cached_property
    def sequence(self) -> List[Union[str, Highlight]]:
        parser = _HighlightParser()
        parser.feed(self.html)
        sequence = parser.sequence
        parser.close()
        return sequence


class _HighlightParser(HTMLParser):
    _sequence: List[Union[str, Highlight]]
    _data: Optional[str]
    _highlight_depth: int

    def __init__(self):
        super().__init__(convert_charrefs=True)
        self._sequence = []
        self._data = None
        self._highlight_depth = 0

    def _flush_data(self):
        if self._data is not None:
            if self._highlight_depth > 0:
                self._sequence.append(
                    Highlight(
                        self._data,
                        depth=self._highlight_depth
                    )
                )
            else:
                self._sequence.append(self._data)
        self._data = None

    @property
    def sequence(self) -> List[Union[str, Highlight]]:
        self._flush_data()
        return self._sequence

    def handle_starttag(self, tag: str, attrs):
        if tag != "em":
            raise SyntaxError("Can only parse <em> tags.")
        if attrs:
            raise SyntaxError("Cannot parse attributes.")
        self._flush_data()
        self._highlight_depth += 1

    # Overridable -- handle end tag
    def handle_endtag(self, tag: str):
        if tag != "em":
            raise SyntaxError("Can only parse <em> tags.")
        self._flush_data()
        self._highlight_depth -= 1

    def handle_charref(self, name: str):
        raise AssertionError(
            "Should never be called because convert_charrefs is True."
        )

    def handle_entityref(self, name: str):
        raise AssertionError(
            "Should never be called because convert_charrefs is True."
        )

    def handle_data(self, data: str):
        if self._data is None:
            self._data = data
        else:
            self._data += data

    def handle_comment(self, data: str):
        raise SyntaxError("Comments are not supported.")

    def handle_decl(self, decl: str):
        raise SyntaxError("Doctype declarations are not supported.")

    def handle_pi(self, data: str):
        raise SyntaxError("Processing instructions are not supported.")



from itertools import chain
from typing import Iterable, Iterator
from uuid import uuid5
from warnings import warn

from click import echo
from elasticsearch_dsl import Search
from elasticsearch_dsl.function import RandomScore
from elasticsearch_dsl.query import FunctionScore, Exists, Term
from tqdm.auto import tqdm

from archive_query_log.config import Config
from archive_query_log.namespaces import NAMESPACE_SOURCE
from archive_query_log.orm import Archive, Provider, Source, InnerArchive, InnerProvider
from archive_query_log.utils.es import safe_iter_scan, update_action
from archive_query_log.utils.time import utc_now


def _sources_batch(archive: Archive, provider: Provider, config: Config) -> list[dict]:
    if provider.exclusion_reason is not None:
        warn(
            f"Skipping provider {provider.id} "
            f"because it is excluded: {provider.exclusion_reason}"
        )
        return []

    batch = []
    for domain in provider.domains:
        for url_path_prefix in provider.url_path_prefixes:
            source_id_components = (
                archive.cdx_api_url,
                archive.memento_api_url,
                domain,
                url_path_prefix,
            )
            source_id = str(
                uuid5(
                    NAMESPACE_SOURCE,
                    ":".join(source_id_components),
                )
            )
            source = Source(
                id=source_id,
                last_modified=utc_now(),
                archive=InnerArchive(
                    id=archive.id,
                    cdx_api_url=archive.cdx_api_url,
                    memento_api_url=archive.memento_api_url,
                    priority=archive.priority,
                ),
                provider=InnerProvider(
                    id=provider.id,
                    domain=domain,
                    url_path_prefix=url_path_prefix,
                    priority=provider.priority,
                ),
                should_fetch_captures=True,
            )
            source.meta.index = config.es.index_sources
            batch.append(source.to_dict(include_meta=True))
    return batch


def _iter_sources_batches_changed_archives(
    changed_archives_search: Search,
    all_providers_search: Search,
    config: Config,
) -> Iterator[list[dict]]:
    archive: Archive
    provider: Provider
    changed_archives = changed_archives_search.scan()
    changed_archives = safe_iter_scan(changed_archives)
    for archive in changed_archives:
        all_providers = all_providers_search.scan()
        all_providers = safe_iter_scan(all_providers)
        for provider in all_providers:
            yield _sources_batch(
                archive,
                provider,
                config,
            )
        yield [
            update_action(
                archive,
                should_build_sources=False,
                last_built_sources=utc_now(),
            )
        ]


def _iter_sources_batches_changed_providers(
    changed_providers_search: Search,
    all_archives_search: Search,
    config: Config,
) -> Iterator[list[dict]]:
    archive: Archive
    provider: Provider
    changed_providers = changed_providers_search.scan()
    changed_providers = safe_iter_scan(changed_providers)
    for provider in changed_providers:
        all_archives = all_archives_search.scan()
        all_archives = safe_iter_scan(all_archives)
        for archive in all_archives:
            yield _sources_batch(
                archive,
                provider,
                config,
            )
        yield [
            update_action(
                provider,
                should_build_sources=False,
                last_built_sources=utc_now(),
            )
        ]


def _build_archive_sources(config: Config) -> None:
    config.es.client.indices.refresh(index=config.es.index_archives)
    config.es.client.indices.refresh(index=config.es.index_providers)
    changed_archives_search = (
        Archive.search(using=config.es.client, index=config.es.index_archives)
        .filter(~Term(should_build_sources=False))
        .query(FunctionScore(functions=[RandomScore()]))
    )
    num_changed_archives = changed_archives_search.count()
    all_providers_search = Provider.search(
        using=config.es.client, index=config.es.index_providers
    ).filter(~Exists(field="exclusion_reason"))
    num_all_providers = all_providers_search.count()
    num_batches_archives = (
        num_changed_archives * num_all_providers
    ) + num_changed_archives
    if num_batches_archives > 0:
        echo(f"Building sources for {num_changed_archives} " f"new/changed archives.")
        action_batches_archives: Iterable[list[dict]] = (
            _iter_sources_batches_changed_archives(
                changed_archives_search=changed_archives_search,
                all_providers_search=all_providers_search,
                config=config,
            )
        )
        # noinspection PyTypeChecker
        action_batches_archives = tqdm(
            action_batches_archives,
            total=num_batches_archives,
            desc="Build sources",
            unit="batch",
        )
        actions_archives = chain.from_iterable(action_batches_archives)
        config.es.bulk(actions_archives)
    else:
        echo("No new/changed archives.")


def _build_provider_sources(config: Config) -> None:
    config.es.client.indices.refresh(index=config.es.index_archives)
    config.es.client.indices.refresh(index=config.es.index_providers)
    changed_providers_search = (
        Provider.search(using=config.es.client, index=config.es.index_providers)
        .filter(~Term(should_build_sources=False))
        .query(FunctionScore(functions=[RandomScore()]))
    )
    num_changed_providers = changed_providers_search.count()
    all_archives_search = Archive.search(
        using=config.es.client, index=config.es.index_archives
    )
    num_all_archives = all_archives_search.count()
    num_batches_providers = (
        num_changed_providers * num_all_archives
    ) + num_changed_providers
    if num_batches_providers > 0:
        echo(f"Building sources for {num_changed_providers} " f"new/changed providers.")
        action_batches_providers: Iterable[list[dict]] = (
            _iter_sources_batches_changed_providers(
                changed_providers_search=changed_providers_search,
                all_archives_search=all_archives_search,
                config=config,
            )
        )
        # noinspection PyTypeChecker
        action_batches_providers = tqdm(
            action_batches_providers,
            total=num_batches_providers,
            desc="Build sources",
            unit="batch",
        )
        actions_providers = chain.from_iterable(action_batches_providers)
        config.es.bulk(actions_providers)
    else:
        echo("No new/changed providers.")


def build_sources(
    config: Config,
    skip_archives: bool,
    skip_providers: bool,
) -> None:
    if not skip_archives:
        _build_archive_sources(config)
    if not skip_providers:
        _build_provider_sources(config)



from itertools import chain, islice
from typing import Iterable, Iterator
from urllib.parse import urljoin
from uuid import uuid5
from warnings import warn

from click import echo
from elasticsearch_dsl import Search
from elasticsearch_dsl.function import RandomScore
from elasticsearch_dsl.query import FunctionScore, RankFeature, Term
from requests import ConnectTimeout, HTTPError, Response
from tqdm.auto import tqdm
from web_archive_api.cdx import CdxApi, CdxMatchType

from archive_query_log.config import Config
from archive_query_log.namespaces import NAMESPACE_CAPTURE
from archive_query_log.orm import Source, Capture, InnerParser
from archive_query_log.utils.es import safe_iter_scan, update_action
from archive_query_log.utils.time import utc_now, UTC


def _iter_captures(
        config: Config,
        source: Source,
) -> Iterator[Capture]:
    cdx_api = CdxApi(
        api_url=source.archive.cdx_api_url,
        session=config.http.session,
    )
    url = f"https://{source.provider.domain}"
    url = urljoin(url, source.provider.url_path_prefix)
    url = url.removeprefix("https://")
    cdx_captures = cdx_api.iter_captures(
        url=url,
        match_type=CdxMatchType.PREFIX,
    )
    for cdx_capture in cdx_captures:
        if len(cdx_capture.url) > 32766:
            warn(RuntimeWarning(
                f"The URL {cdx_capture.url} exceeds the "
                f"maximum length of Elasticsearch."
                f" It will be skipped."
            ))
            continue

        capture_utc_timestamp_text = (
            cdx_capture.timestamp.astimezone(UTC).strftime("%Y%m%d%H%M%S"))
        capture_id_components = (
            source.archive.cdx_api_url,
            cdx_capture.url,
            capture_utc_timestamp_text,
        )
        capture_id = str(uuid5(
            NAMESPACE_CAPTURE,
            ":".join(capture_id_components),
        ))
        yield Capture(
            id=capture_id,
            last_modified=utc_now(),
            archive=source.archive,
            provider=source.provider,
            url=cdx_capture.url,
            url_key=cdx_capture.url_key,
            timestamp=cdx_capture.timestamp.astimezone(UTC),
            status_code=cdx_capture.status_code,
            digest=cdx_capture.digest,
            mimetype=cdx_capture.mimetype,
            filename=cdx_capture.filename,
            offset=cdx_capture.offset,
            length=cdx_capture.length,
            access=cdx_capture.access,
            redirect_url=cdx_capture.redirect_url,
            flags=([flag.value for flag in cdx_capture.flags]
                   if cdx_capture.flags is not None else None),
            collection=cdx_capture.collection,
            source=cdx_capture.source,
            source_collection=cdx_capture.source_collection,
            url_query_parser=InnerParser(
                should_parse=True,
            ),
        )


def _add_captures_actions(
        config: Config,
        source: Source,
) -> Iterator[dict]:
    # Re-check if fetching captures is necessary.
    if (source.should_fetch_captures is not None and
            not source.should_fetch_captures):
        return

    captures_iter = _iter_captures(config, source)
    try:
        for capture in captures_iter:
            capture.meta.index = config.es.index_captures
            yield capture.to_dict(include_meta=True)
    except ConnectTimeout as e:
        # The archives' CDX are usually very slow, so we expect timeouts.
        # Rather than failing, we just warn and continue with the next source.
        # But we do not mark this source as fetched, so that we try again.
        warn(RuntimeWarning(
            f"Connection timeout while fetching captures "
            f"for source {source.id}: {e}"))
        return
    except HTTPError as e:
        ignored = False
        response: Response = e.response
        if response.status_code == 403:
            warn(RuntimeWarning(
                f"Unauthorized to fetch captures for source "
                f"domain {source.provider.domain} and "
                f"URL prefix {source.provider.url_path_prefix}."
            ))
            ignored = True
        if not ignored:
            raise e

    yield update_action(
        source,
        should_fetch_captures=False,
        last_fetched_captures=utc_now(),
    )


def fetch_captures(config: Config, prefetch_limit: int | None = None) -> None:
    changed_sources_search: Search = (
        Source.search(using=config.es.client, index=config.es.index_sources)
        .filter(~Term(should_fetch_captures=False))
        .query(
            RankFeature(field="archive.priority", saturation={}) |
            RankFeature(field="provider.priority", saturation={}) |
            FunctionScore(functions=[RandomScore()])
        )
    )
    num_changed_sources = changed_sources_search.count()
    if num_changed_sources > 0:
        echo(f"Fetching captures for {num_changed_sources} "
             f"new/changed sources.")
        changed_sources: Iterable[Source] = (
            changed_sources_search
            .params(preserve_order=True)
            .scan()
        )
        changed_sources = safe_iter_scan(changed_sources)

        if prefetch_limit is not None:
            num_changed_sources = min(num_changed_sources, prefetch_limit)
            changed_sources = tqdm(changed_sources, total=num_changed_sources, desc="Pre-fetching sources", unit="source")
            changed_sources = iter(list(islice(changed_sources, prefetch_limit)))

        # noinspection PyTypeChecker
        changed_sources = tqdm(changed_sources, total=num_changed_sources,
                               desc="Fetching captures", unit="source")
        actions = chain.from_iterable(
            _add_captures_actions(config, source)
            for source in changed_sources
        )
        config.es.bulk(actions)
    else:
        echo("No new/changed sources.")



from io import BytesIO
from shutil import copyfileobj
from warnings import warn

from warcio.recordloader import ArcWarcRecord


def read_html_string(record: ArcWarcRecord) -> str | None:
    mime_type: str | None = record.http_headers.get_header("Content-Type")
    if mime_type is None:
        warn(UserWarning("No MIME type given."))
        return None
    mime_type = mime_type.split(";", maxsplit=1)[0]
    if mime_type != "text/xml":
        return None
    with BytesIO() as content_buffer:
        copyfileobj(record.content_stream(), content_buffer)
        return content_buffer.getvalue().decode("utf-8")



from contextlib import asynccontextmanager
from typing import AsyncIterator

from aiohttp import ClientSession, TCPConnector, ClientTimeout, \
    ClientConnectorError, ServerTimeoutError, ClientPayloadError
from aiohttp_retry import RetryClient, JitterRetry


@asynccontextmanager
async def archive_http_session(
        limit: int = 10,
) -> AsyncIterator[ClientSession]:
    # The Wayback Machine doesn't seem to support more than 10
    # parallel connections from the same IP.
    connector = TCPConnector(
        limit_per_host=limit,
    )
    # Graceful timeout as the Wayback Machine is sometimes very slow.
    timeout = ClientTimeout(
        total=15 * 60,
        connect=5 * 60,  # Setting up a connection is especially slow.
        sock_read=5 * 60,
    )
    async with ClientSession(
            connector=connector,
            timeout=timeout,
    ) as session:
        yield session


@asynccontextmanager
async def archive_http_client(limit: int = 10) -> AsyncIterator[RetryClient]:
    retry_options = JitterRetry(
        attempts=10,
        start_timeout=10,  # 10 seconds
        max_timeout=15 * 60,  # 15 minutes
        statuses={429, 502, 503, 504},  # server errors
        exceptions={
            ClientConnectorError,
            ServerTimeoutError,
            ClientPayloadError,
        },
    )
    async with archive_http_session(limit) as session:
        retry_client = RetryClient(
            client_session=session,
            retry_options=retry_options,
        )
        yield retry_client



from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry


def backoff_session() -> Session:
    session = Session()
    retries = Retry(
        total=10,
        backoff_factor=1,
        status_forcelist=[502, 503, 504],
    )
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session



from datetime import datetime
from functools import cached_property
from re import Pattern, compile as pattern
from typing import Literal

from elasticsearch_dsl import Document, Keyword, Text, Date, RankFeature, \
    InnerDoc as InnerDocument, Object, Index, Integer, Nested, Long, Boolean


class BaseDocument(Document):
    last_modified: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )

    # TODO: At the moment, this is used more as a creation date.
    #  We could use a different field for that and use this one for the last
    #  modified date.

    # noinspection PyShadowingBuiltins
    def __init__(self, id: str | None = None, **kwargs):
        if id is not None:
            if "meta" not in kwargs:
                kwargs["meta"] = {}
            kwargs["meta"]["id"] = id
        super().__init__(**kwargs)

    @classmethod
    def index(cls) -> Index:
        return cls._index

    @property
    def id(self) -> str:
        return self.meta.id

    @id.setter
    def id(self, value: str):
        self.meta.id = value


class Archive(BaseDocument):
    name: str = Text()
    description: str = Text()
    cdx_api_url: str = Keyword()
    memento_api_url: str = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    should_build_sources: bool = Boolean()
    last_built_sources: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


class Provider(BaseDocument):
    name: str = Text()
    description: str = Text()
    exclusion_reason: str = Text()
    notes: str = Text()
    domains: list[str] = Keyword()
    url_path_prefixes: list[str] = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    should_build_sources: bool = Boolean()
    last_built_sources: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


class InnerArchive(InnerDocument):
    id: str = Keyword()
    cdx_api_url: str = Keyword()
    memento_api_url: str = Keyword()
    priority: int | None = RankFeature(positive_score_impact=True)


class InnerProvider(InnerDocument):
    id: str = Keyword()
    domain: str = Keyword()
    url_path_prefix: str = Keyword()
    priority: int | None = RankFeature(positive_score_impact=True)


class Source(BaseDocument):
    archive: InnerArchive = Object(InnerArchive)
    provider: InnerProvider = Object(InnerProvider)
    should_fetch_captures: bool = Boolean()
    last_fetched_captures: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )

    class Index:
        settings = {
            "number_of_shards": 5,
            "number_of_replicas": 2,
        }


class InnerParser(InnerDocument):
    id: str = Keyword()
    should_parse: bool = Boolean()
    last_parsed: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )


class Capture(BaseDocument):
    archive: InnerArchive = Object(InnerArchive)
    provider: InnerProvider = Object(InnerProvider)
    url: str = Keyword()
    url_key: str = Keyword()
    timestamp: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )
    status_code: int = Integer()
    digest: str = Keyword()
    mimetype: str | None = Keyword()
    filename: str | None = Keyword()
    offset: int | None = Integer()
    length: int | None = Integer()
    access: str | None = Keyword()
    redirect_url: str | None = Keyword()
    flags: list[str] | None = Keyword()
    collection: str | None = Keyword()
    source: str | None = Keyword()
    source_collection: str | None = Keyword()
    url_query_parser: InnerParser | None = Object(InnerParser)

    class Index:
        settings = {
            "number_of_shards": 40,
            "number_of_replicas": 2,
        }


class InnerCapture(InnerDocument):
    id: str = Keyword()
    url: str = Keyword()
    timestamp: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )
    status_code: int = Integer()
    digest: str = Keyword()
    mimetype: str | None = Keyword()


class InnerDownloader(InnerDocument):
    id: str = Keyword()
    should_download: bool = Boolean()
    last_downloaded: datetime = Date(
        default_timezone="UTC",
        format="strict_date_time_no_millis",
    )


class WarcLocation(InnerDocument):
    file: str = Keyword()
    offset: int = Long()
    length: int = Long()


class SnippetId(InnerDocument):
    id: str = Keyword()
    rank: int = Integer()


class Snippet(SnippetId):
    content: str = Text()
    url: str | None = Keyword()
    title: str | None = Text()
    text: str | None = Text()


class DirectAnswerId(InnerDocument):
    id: str = Keyword()


class DirectAnswer(DirectAnswerId):
    content: str = Text()
    url: str | None = Keyword()
    text: str | None = Text()


class Serp(BaseDocument):
    archive: InnerArchive = Object(InnerArchive)
    provider: InnerProvider = Object(InnerProvider)
    capture: InnerCapture = Object(InnerCapture)
    url_query: str = Text()
    url_query_parser: InnerParser | None = Object(InnerParser)
    url_page: int | None = Integer()
    url_page_parser: InnerParser | None = Object(InnerParser)
    url_offset: int | None = Integer()
    url_offset_parser: InnerParser | None = Object(InnerParser)
    # url_language: str | None = Keyword()
    # url_language_parser: InnerParser | None = Object(InnerParser)
    warc_location: WarcLocation | None = Object(WarcLocation)
    warc_downloader: InnerDownloader | None = Object(InnerDownloader)
    warc_query: str | None = Text()
    warc_query_parser: InnerParser | None = Object(InnerParser)
    warc_snippets: list[SnippetId] | None = Nested(SnippetId)
    warc_snippets_parser: InnerParser | None = Object(InnerParser)
    warc_direct_answers: list[DirectAnswerId] | None = Nested(DirectAnswerId)
    warc_direct_answers_parser: InnerParser | None = Object(InnerParser)

    # rendered_warc_location: WarcLocation | None = Object(WarcLocation)
    # rendered_warc_downloader: InnerDownloader | None = (
    #     Object(InnerDownloader))

    class Index:
        settings = {
            "number_of_shards": 40,
            "number_of_replicas": 2,
        }


class InnerSerp(InnerDocument):
    id: str = Keyword()


class Result(BaseDocument):
    archive: InnerArchive = Object(InnerArchive)
    provider: InnerProvider = Object(InnerProvider)
    capture: InnerCapture = Object(InnerCapture)
    serp: InnerSerp = Object(InnerSerp)
    snippet: Snippet = Object(Snippet)
    snippet_parser: InnerParser | None = Object(InnerParser)
    warc_before_serp_location: WarcLocation | None = Object(WarcLocation)
    warc_before_serp_downloader: InnerDownloader | None = (
        Object(InnerDownloader))
    warc_after_serp_location: WarcLocation | None = Object(WarcLocation)
    warc_after_serp_downloader: InnerDownloader | None = (
        Object(InnerDownloader))

    class Index:
        settings = {
            "number_of_shards": 20,
            "number_of_replicas": 2,
        }


class InnerProviderId(InnerDocument):
    id: str = Keyword()


UrlQueryParserType = Literal[
    "query_parameter",
    "fragment_parameter",
    "path_segment",
]


class UrlQueryParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: UrlQueryParserType = Keyword()
    parameter: str | None = Keyword()
    segment: int | None = Keyword()
    remove_pattern_regex: str | None = Keyword()
    space_pattern_regex: str | None = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    @cached_property
    def remove_pattern(self) -> Pattern | None:
        if self.remove_pattern_regex is None:
            return None
        return pattern(self.remove_pattern_regex)

    @cached_property
    def space_pattern(self) -> Pattern | None:
        if self.space_pattern_regex is None:
            return None
        return pattern(self.space_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


UrlPageParserType = Literal[
    "query_parameter",
    "fragment_parameter",
    "path_segment",
]


class UrlPageParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: UrlPageParserType = Keyword()
    parameter: str | None = Keyword()
    segment: int | None = Keyword()
    remove_pattern_regex: str | None = Keyword()
    space_pattern_regex: str | None = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    @cached_property
    def remove_pattern(self) -> Pattern | None:
        if self.remove_pattern_regex is None:
            return None
        return pattern(self.remove_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


UrlOffsetParserType = Literal[
    "query_parameter",
    "fragment_parameter",
    "path_segment",
]


class UrlOffsetParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: UrlOffsetParserType = Keyword()
    parameter: str | None = Keyword()
    segment: int | None = Keyword()
    remove_pattern_regex: str | None = Keyword()
    space_pattern_regex: str | None = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    @cached_property
    def remove_pattern(self) -> Pattern | None:
        if self.remove_pattern_regex is None:
            return None
        return pattern(self.remove_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


WarcQueryParserType = Literal[
    "xpath",
]


class WarcQueryParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: WarcQueryParserType = Keyword()
    xpath: str | None = Keyword()
    remove_pattern_regex: str | None = Keyword()
    space_pattern_regex: str | None = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    @cached_property
    def remove_pattern(self) -> Pattern | None:
        if self.remove_pattern_regex is None:
            return None
        return pattern(self.remove_pattern_regex)

    @cached_property
    def space_pattern(self) -> Pattern | None:
        if self.space_pattern_regex is None:
            return None
        return pattern(self.space_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


WarcSnippetsParserType = Literal[
    "xpath",
]


class WarcSnippetsParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: WarcSnippetsParserType = Keyword()
    xpath: str | None = Keyword()
    url_xpath: str | None = Keyword()
    title_xpath: str | None = Keyword()
    text_xpath: str | None = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


WarcDirectAnswersParserType = Literal[
    "xpath",
]


class WarcDirectAnswersParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: WarcDirectAnswersParserType = Keyword()
    xpath: str | None = Keyword()
    url_xpath: str | None = Keyword()
    text_xpath: str | None = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }


WarcMainContentParserType = Literal[
    "resiliparse",
]


class WarcMainContentParser(BaseDocument):
    provider: InnerProviderId | None = Object(InnerProviderId)
    url_pattern_regex: str | None = Keyword()
    priority: float | None = RankFeature(positive_score_impact=True)
    parser_type: WarcMainContentParserType = Keyword()

    @cached_property
    def url_pattern(self) -> Pattern | None:
        if self.url_pattern_regex is None:
            raise ValueError("No URL pattern regex.")
        return pattern(self.url_pattern_regex)

    class Index:
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 2,
        }



from bleach import clean
from bs4 import Tag

_HIGHLIGHT_TAGS = ["em", "strong", "mark", "b", "i", "u"]


def clean_html(html: str | Tag) -> str:
    if isinstance(html, Tag):
        html = html.decode_contents()
    html = clean(
        html,
        tags=_HIGHLIGHT_TAGS,
        attributes=[],
        protocols=[],
        strip=True,
        strip_comments=True,
    )
    for tag in _HIGHLIGHT_TAGS:
        html = html.replace(f"<{tag}>", "<em>")
        html = html.replace(f"</{tag}>", "</em>")
    html = html.strip()
    return html



from approvaltests import set_default_reporter, DiffReporter
from pytest import fixture


def configure_approvaltests():
    set_default_reporter(DiffReporter())


@fixture(scope="session", autouse=True)
def set_default_reporter_for_all_tests():
    configure_approvaltests()



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_360_an_jian_diao_cha_bi_lu_1576497545():
    verify_serp_parsing(
        "https://web.archive.org/web/20191216115905id_/https://www.so.com/s?q=%E6%A1%88%E4%BB%B6%E8%B0%83%E6%9F%A5%E7%AC%94%E5%BD%95&src=related_2.6&psid=836a2b7f45bb7fe2b7bcbe6bbe47b867&nlpv=zzzc_base_2",
        "360",
    )


def test_parse_query_360_chen_teng_tan_han_han_mei_ren_jie_ji_1577810108():
    verify_serp_parsing(
        "https://web.archive.org/web/20191231163508id_/https://www.so.com/s?ie=utf-8&src=know_side_nlp_sohot_online&q=%E6%B2%88%E8%85%BE%E8%B0%88%E9%9F%A9%E5%AF%92%E6%B2%A1%E4%BA%BA%E6%8E%A5%E6%9C%BA&ob_ext=%7B%22rsv_cq%22%3A%22%5Cu514d%5Cu8d39vpn%22%2C%22rt%22%3A%22%5Cu5b9e%5Cu65f6%5Cu70ed%5Cu641c%22%2C%22rclken%22%3Anull%7D",
        "360",
    )


def test_parse_query_360_jin_ji_che_hui_1576637402():
    verify_serp_parsing(
        "https://web.archive.org/web/20191218025002id_/https://www.so.com/s?q=%E7%B4%A7%E6%80%A5%E6%92%A4%E5%9B%9E&src=webcache",
        "360",
    )


def test_parse_query_360_shi_jie_zui_ai_de_ren_qu_shi_1579483438():
    verify_serp_parsing(
        "https://web.archive.org/web/20200120012358id_/https://www.so.com/s?ie=utf-8&src=know_side_nlp_sohot_online&q=%E4%B8%96%E7%95%8C%E6%9C%80%E7%9F%AE%E7%9A%84%E4%BA%BA%E5%8E%BB%E4%B8%96&ob_ext=%7B%22rsv_cq%22%3A%22openeuler%22%2C%22rt%22%3A%22%5Cu5b9e%5Cu65f6%5Cu70ed%5Cu641c%22%2C%22rclken%22%3Anull%7D",
        "360",
    )


def test_parse_query_360_21tian_jian_fei_fa_1653652529():
    verify_serp_parsing(
        "https://web.archive.org/web/20220527115529id_/https://www.so.com/s?src=lm&ls=s4b0c48ca9b&q=21%E5%A4%A9%E5%87%8F%E8%82%A5%E6%B3%95&lmsid=af298c70a1d01987&lm_extend=ctype%3A20%7Clmbid%3A200%7Cjt%3A2%7Cmaxbid%3A%7Csadspace%3A%7Cadid%3A3444654665",
        "360",
    )


def test_parse_query_360_site_chaxun_biz_1551742698():
    verify_serp_parsing(
        "https://web.archive.org/web/20190304233818id_/https://www.so.com/s?q=site:chaxun.biz",
        "360",
    )


def test_parse_query_360_si_chuan_fa_sheng_6_8ji_di_zhen_1662384982():
    verify_serp_parsing(
        "https://web.archive.org/web/20220905133622id_/https://www.so.com/s?q=%E5%9B%9B%E5%B7%9D%E5%8F%91%E7%94%9F6.8%E7%BA%A7%E5%9C%B0%E9%9C%87&src=bkrc&srcg=&stat_source=sshot",
        "360",
    )


def test_parse_query_360_hui_ji_pei_xun_1579668897():
    verify_serp_parsing(
        "https://web.archive.org/web/20200122045457id_/https://www.so.com/s?src=lm&ls=s521c48db9a&q=%E4%BC%9A%E8%AE%A1%E5%9F%B9%E8%AE%AD&lmsid=59a53f2a01747c88&lm_extend=ctype%3A4%7Clmbid%3A20%2C2%2C37%2C30%2C64%2C74%2C7%2C93%2C101%2C110%2C500%7Cjt%3A2%7Cmaxbid%3A%7Csadspace%3A",
        "360",
    )


def test_parse_query_360_site_chaxun_biz_1550161171():
    verify_serp_parsing(
        "https://web.archive.org/web/20190214161931id_/https://www.so.com/s?q=site:chaxun.biz",
        "360",
    )


def test_parse_query_360_mao_bu_yi_xiao_chou_chao_xi_1576791478():
    verify_serp_parsing(
        "https://web.archive.org/web/20191219213758id_/https://www.so.com/s?q=%E6%AF%9B%E4%B8%8D%E6%98%93%E6%B6%88%E6%84%81%E6%8A%84%E8%A2%AD&src=related_2.6&psid=ce3a7e94b02d7f56b85d23614a2be0ae&nlpv=mobile_cooccur_v1",
        "360",
    )


def test_parse_query_360_mei_guo_qian_zheng_1579954288():
    verify_serp_parsing(
        "https://web.archive.org/web/20200125121128id_/https://www.so.com/s?src=lm&ls=s521c48db9a&q=%E7%BE%8E%E5%9B%BD%E7%AD%BE%E8%AF%81&lmsid=30e88c9397a90e32&lm_extend=ctype%3A4%7Clmbid%3A203%7Cjt%3A2%7Cmaxbid%3A%7Csadspace%3A",
        "360",
    )


def test_parse_query_360_duo_jia_ju_chang_qu_xiao_yan_chu_1579663841():
    verify_serp_parsing(
        "https://web.archive.org/web/20200122033041id_/https://www.so.com/s?ie=utf-8&src=know_side_nlp_sohot_online&q=%E5%A4%9A%E5%AE%B6%E5%89%A7%E5%9C%BA%E5%8F%96%E6%B6%88%E6%BC%94%E5%87%BA&ob_ext=%7B%22rsv_cq%22%3A%22%5Cu8003%5Cu7814%5Cu4f5c%5Cu5f0a%22%2C%22rt%22%3A%22%5Cu5b9e%5Cu65f6%5Cu70ed%5Cu641c%22%2C%22rclken%22%3Anull%7D",
        "360",
    )


def test_parse_query_360_wu_xiu_bo_yang_kun_ju_hui_hai_ge_1577858222():
    verify_serp_parsing(
        "https://web.archive.org/web/20200101055702id_/https://www.so.com/s?ie=utf-8&src=know_side_nlp_sohot_online&q=%E5%90%B4%E7%A7%80%E6%B3%A2%E6%9D%A8%E5%9D%A4%E8%81%9A%E4%BC%9A%E5%97%A8%E6%AD%8C&ob_ext=%7B%22rsv_cq%22%3A%22%5Cu4e2a%5Cu4eba%5Cu6570%5Cu636e%5Cu6cc4%5Cu9732%22%2C%22rt%22%3A%22%5Cu5b9e%5Cu65f6%5Cu70ed%5Cu641c%22%2C%22rclken%22%3Anull%7D",
        "360",
    )


def test_parse_query_360_tao_bao_ru_he_che_hui_ping_jie_1576604187():
    verify_serp_parsing(
        "https://web.archive.org/web/20191217173627id_/https://www.so.com/s?q=%E6%B7%98%E5%AE%9D%E5%A6%82%E4%BD%95%E6%92%A4%E5%9B%9E%E8%AF%84%E4%BB%B7&psid=ddf2765c1c51375b28acdb33c3767871&nlpv=zzzc_base_1&src=pdr_guide_2.6",
        "360",
    )


def test_parse_query_360_site_chaxun_biz_1550145271():
    verify_serp_parsing(
        "https://web.archive.org/web/20190214115431id_/https://www.so.com/s?q=site:chaxun.biz",
        "360",
    )


def test_parse_query_360_ju_min_jia_zhong_fei_jin_bian_fu_1585788499():
    verify_serp_parsing(
        "https://web.archive.org/web/20200402004819id_/https://www.so.com/s?q=%E5%B1%85%E6%B0%91%E5%AE%B6%E4%B8%AD%E9%A3%9E%E8%BF%9B%E8%9D%99%E8%9D%A0&src=know_side_nlp_sohot&tn=news",
        "360",
    )


def test_parse_query_360_site_eng_kaz_enacademic_com_1581787441():
    verify_serp_parsing(
        "https://web.archive.org/web/20200215172401id_/https://www.so.com/s?q=site%3Aeng_kaz.enacademic.com",
        "360",
    )


def test_parse_query_360_niu_jie_5hao_1576819311():
    verify_serp_parsing(
        "https://web.archive.org/web/20191220052151id_/https://www.so.com/s?q=%E7%89%9B%E6%9D%B05%E5%8F%B7&src=related_2.6&psid=7023274fd3a925bfe6b5bc6c27debf13&nlpv=zzzc_base_2",
        "360",
    )


def test_parse_query_360_zhang_bo_zhi_bei_gou_yao_shang_1661090641():
    verify_serp_parsing(
        "https://web.archive.org/web/20220821140401id_/https://www.so.com/s?ie=utf-8&src=know_side_nlp_sohot_online_clk&q=%E5%BC%A0%E6%9F%8F%E8%8A%9D%E8%A2%AB%E7%8B%97%E5%92%AC%E4%BC%A4&ob_ext=%7B%22rsv_cq%22%3A%22%5Cu6bdb%5Cu4e0d%5Cu6613%22%2C%22rt%22%3A%22%5Cu5b9e%5Cu65f6%5Cu70ed%5Cu641c%22%2C%22rclken%22%3Anull%7D",
        "360",
    )


def test_parse_query_360_lu_xing_she_tuan_dui_you_zan_ting_1580014331():
    verify_serp_parsing(
        "https://web.archive.org/web/20200126045211id_/https://www.so.com/s?ie=utf-8&q=%E6%97%85%E8%A1%8C%E7%A4%BE%E5%9B%A2%E9%98%9F%E6%B8%B8%E6%9A%82%E5%81%9C&src=no_result_hot",
        "360",
    )



# flake8: noqa
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_9_11_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210411015549id_/http://www.google.com/search?hl=en&lr=&ie=ISO-8859-1&q=%22+9/11+revisited%22',
        'google'
    )


def test_lenin_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210227002632id_/http://www.google.com/search?hl=en&lr=&ie=ISO-8859-1&q=%22+Lenin%22',
        'google'
    )


def test_august_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210225064449id_/http://www.google.com/search?hl=en&q=%22%22+%22Agust&iacute;n',
        'google'
    )


def test_cortisol_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210609022746id_/http://www.google.com/search?hl=en&lr=&ie=ISO-8859-1&safe=off&q=++++%22+Cortisol+test%22',
        'google'
    )


def test_coxsackie_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210330123639id_/http://www.google.com/search?hl=en&ie=ISO-8859-1&q=%22+coxsackie+virus%22_',
        'google'
    )


def test_homemade_l_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210330123719id_/http://www.google.com/search?hl=en&source=hp&ie=ISO-8859-1&q=%22+homemade+l',
        'google'
    )


def test_homemade_dove_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210330123611id_/http://www.google.com/search?hl=en&ie=ISO-8859-1&q=%22+dove%22+%22soap%22+%22',
        'google'
    )


def test_dead_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20210224214643id_/http://www.google.com/search?hl=en&lr=&ie=ISO-8859-1&q=%22+dead+cock%22+mortuary',
        'google'
    )


def test_boston_hotels_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20030829021921id_/http://www.google.com:80/search?q=%22+www.boston-hotels-cheap.net%22',
        'google'
    )


def test_extractresult_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20030831234927id_/http://www.google.com:80/search?q=%22+www.exactresult.+com/%22',
        'google'
    )


def test_wallstquotes_search():
    verify_serp_parsing(
        'https://web.archive.org/web/20030829070453id_/http://www.google.com:80/search?q=%22+www.wallstquotes.+com%22',
        'google'
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_stackoverflow_objective_c_1354546520():
    verify_serp_parsing(
        "https://web.archive.org/web/20121203145520id_/http://stackoverflow.com/questions/tagged/objective-c",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_scala_1355718931():
    verify_serp_parsing(
        "https://web.archive.org/web/20121217043531id_/http://stackoverflow.com/questions/tagged/scala",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_remote_execution_1645839151():
    verify_serp_parsing(
        "https://web.archive.org/web/20220226013231id_/https://stackoverflow.com/questions/tagged/remote-execution",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_text_mining_1522322354():
    verify_serp_parsing(
        "https://web.archive.org/web/20180329111914id_/https://stackoverflow.com/questions/tagged/text-mining",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_vue_js_php_1647710017():
    verify_serp_parsing(
        "https://web.archive.org/web/20220319171337id_/https://stackoverflow.com/questions/tagged/vue.js+php",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_ruby_on_rails_plugins_1613971498():
    verify_serp_parsing(
        "https://web.archive.org/web/20210222052458id_/https://stackoverflow.com/questions/tagged/ruby-on-rails-plugins",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_ruby_1_9_3_heroku_1398351399():
    verify_serp_parsing(
        "https://web.archive.org/web/20140424145639id_/http://stackoverflow.com/questions/tagged/ruby-1.9.3+heroku",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_fonts_swing_jtextpane_1412916125():
    verify_serp_parsing(
        "https://web.archive.org/web/20141010044205id_/http://stackoverflow.com/questions/tagged/fonts+swing+jtextpane",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_xampp_java_1547581001():
    verify_serp_parsing(
        "https://web.archive.org/web/20190115193641id_/https://stackoverflow.com/questions/tagged/xampp+java",
        "stackoverflow",
    )


def test_parse_query_stackoverflow_numpy_einsum_1516773912():
    verify_serp_parsing(
        "https://web.archive.org/web/20180124060512id_/https://stackoverflow.com/questions/tagged/numpy-einsum",
        "stackoverflow",
    )



from logging import getLogger, Logger
from pathlib import Path

PROJECT_DIRECTORY_PATH: Path = Path(__file__).parent.parent.parent
DATA_DIRECTORY_PATH: Path = PROJECT_DIRECTORY_PATH / "data"

CDX_API_URL = "https://web.archive.org/cdx/search/cdx"

LOGGER: Logger = getLogger(__name__)

if __name__ == "__main__":
    from archive_query_log.legacy.cli import main

    main()
